@@ -375,28 +375,36 @@ int medusaReceiveInternal(int socket, char *buf, int length)
 {
 #ifdef HAVE_LIBSSL
   int err;
+  int nRet;
   struct SSLSOCKETINFO *s;
 
   HASH_FIND_INT( psSSLSocketInfo, &socket, s );
   if ((s != NULL) && (s->nUseSSL))
   {
-    int nRet = SSL_read(s->ssl, buf, length);
-    
-    if (nRet <= 0)
+    do
     {
-      err = SSL_get_error(s->ssl, nRet);
-      switch(err)
+      nRet = SSL_read(s->ssl, buf, length);
+
+      if (nRet <= 0)
       {
-        case SSL_ERROR_ZERO_RETURN:
-          writeError(ERR_ERROR, "The TLS/SSL connection has been closed.");
-          break;
-        case SSL_ERROR_SSL:
-          writeError(ERR_ERROR, "A failure in the SSL library occurred, usually a protocol error.");
-          break;
-        default:
-          writeError(ERR_ERROR, "Unknown TLS/SSL error occurred (%d - %s)", err, ERR_error_string(err, NULL));
+        err = SSL_get_error(s->ssl, nRet);
+        switch(err)
+        {
+          case SSL_ERROR_ZERO_RETURN:
+            writeError(ERR_ERROR, "The TLS/SSL connection has been closed.");
+            break;
+          case SSL_ERROR_SSL:
+            writeError(ERR_ERROR, "A failure in the SSL library occurred, usually a protocol error.");
+            break;
+          case SSL_ERROR_SYSCALL:
+            writeError(ERR_DEBUG, "TLS/SSL I/O error occurered (%d - %s)", err, ERR_error_string(err, NULL));
+            break;
+          default:
+            writeError(ERR_ERROR, "Unknown TLS/SSL error occurred (%d - %s)", err, ERR_error_string(err, NULL));
+        }
       }
     }
+    while (nRet == -1 && err == SSL_ERROR_SYSCALL && errno == EINTR);
 
     return nRet;
   }
@@ -469,12 +477,14 @@ char* medusaReceiveDataInternal(int socket, int* nBufferSize, int nConvertNullsT
    
     if (nBufReceive + nBufReceiveTmp > BUFFER_SIZE) 
     {
-      writeError(ERR_DEBUG, "Additional data received. Increasing receive buffer %d bytes.", nBufReceiveTmp);
+      writeError(ERR_DEBUG, "Additional data received. Increasing receive buffer %d bytes to %d.", nBufReceiveTmp, nBufReceive + nBufReceiveTmp + 1);
       szBufReceive = realloc(szBufReceive, nBufReceive + nBufReceiveTmp + 1);
     }
+
     memcpy(szBufReceive + nBufReceive, szBufReceiveTmp, nBufReceiveTmp);
     nBufReceive += nBufReceiveTmp;
 
+    nBufReceiveTmp = 0;
     free(szBufReceiveTmp);
   }
 
