@@ -23,7 +23,7 @@
  *
 */
 
-#define VERSION_SVN "$Id: medusa.c 603 2006-10-18 20:20:10Z jmk $" 
+#define VERSION_SVN "$Id: medusa.c 1068 2009-02-11 18:59:28Z reefman $" 
 
 #include <dlfcn.h>
 #include "medusa.h"
@@ -34,6 +34,7 @@ char* szTempModuleParam;
 char* szModulePaths[3] = {"a", "b", "c"};         // will look at 3 different locations for modules if possible
 char** arrModuleParams;    // the "argv" for the module
 int nModuleParamCount;    // the "argc" for the module
+int ctrlc = 0;
 
 void freeModuleParams()
 {
@@ -84,6 +85,7 @@ void usage()
   writeVerbose(VB_NONE, "  -v [NUM]     : Verbose level [0 - 6 (more)]");
   writeVerbose(VB_NONE, "  -w [NUM]     : Error debug level [0 - 10 (more)]");
   writeVerbose(VB_NONE, "  -V           : Display version");
+  writeVerbose(VB_NONE, "  -Z [NUM]     : Resume scan from host #");
   writeVerbose(VB_NONE, "\n");
   return;
 }
@@ -110,6 +112,7 @@ int checkOptions(int argc, char **argv, sAudit *_psAudit)
   _psAudit->iRetryWait = WAIT_BETWEEN_CONNECT_RETRY;  /* Default wait of 3 seconds */
   _psAudit->iRetries = MAX_CONNECT_RETRY;             /* Default of 2 retries (3 total attempts) */
   _psAudit->iShowModuleHelp = 0;
+  _psAudit->iResume = 0;
   iVerboseLevel = 5;
   iErrorLevel = 5;
 
@@ -125,14 +128,19 @@ int checkOptions(int argc, char **argv, sAudit *_psAudit)
   if (nIgnoreBanner == 0)
     writeVerbose(VB_NONE, "%s v%s [%s] (C) %s %s\n", PROGRAM, VERSION, WWW, AUTHOR, EMAIL);
 
-  while ((opt = getopt(argc, argv, "h:H:u:U:p:P:C:O:e:M:m:g:r:R:t:T:n:bqdsLfFVv:w:")) != EOF)
+  while ((opt = getopt(argc, argv, "h:H:u:U:p:P:C:O:e:M:m:g:r:R:t:T:n:bqdsLfFVv:w:Z:")) != EOF)
   {
     switch (opt)
     {
     case 'h':
       if (_psAudit->HostType)
       {
-        writeError(ERR_ALERT, "Options 'h' and 'H' are exclusive.");
+        writeError(ERR_ALERT, "Options 'h' and 'H' are mutually exclusive.");
+        ret = EXIT_FAILURE;
+      }
+      if (_psAudit->iResume > 1)
+      {
+        writeError(ERR_ALERT, "You cannot resume a single host scan.");
         ret = EXIT_FAILURE;
       }
       else
@@ -146,7 +154,7 @@ int checkOptions(int argc, char **argv, sAudit *_psAudit)
     case 'H':
       if (_psAudit->HostType)
       {
-        writeError(ERR_ALERT, "Options 'h' and 'H' are exclusive.");
+        writeError(ERR_ALERT, "Options 'h' and 'H' are mutually exclusive.");
         ret = EXIT_FAILURE;
       }
       else
@@ -160,7 +168,7 @@ int checkOptions(int argc, char **argv, sAudit *_psAudit)
     case 'u':
       if (_psAudit->UserType)
       {
-        writeError(ERR_ALERT, "Options 'u' and 'U' are exclusive.");
+        writeError(ERR_ALERT, "Options 'u' and 'U' are mutually exclusive.");
         ret = EXIT_FAILURE;
       }
       else
@@ -175,7 +183,7 @@ int checkOptions(int argc, char **argv, sAudit *_psAudit)
     case 'U':
       if (_psAudit->UserType)
       {
-        writeError(ERR_ALERT, "Options 'u' and 'U' are exclusive.");
+        writeError(ERR_ALERT, "Options 'u' and 'U' are mutually exclusive.");
         ret = EXIT_FAILURE;
       }
       else
@@ -189,7 +197,7 @@ int checkOptions(int argc, char **argv, sAudit *_psAudit)
     case 'p':
       if (_psAudit->PassType)
       {
-        writeError(ERR_ALERT, "Options 'p' and 'P' are exclusive.");
+        writeError(ERR_ALERT, "Options 'p' and 'P' are mutually exclusive.");
         ret = EXIT_FAILURE;
       }
       else
@@ -204,7 +212,7 @@ int checkOptions(int argc, char **argv, sAudit *_psAudit)
     case 'P':
       if (_psAudit->PassType)
       {
-        writeError(ERR_ALERT, "Options 'p' and 'P' are exclusive.");
+        writeError(ERR_ALERT, "Options 'p' and 'P' are mutually exclusive.");
         ret = EXIT_FAILURE;
       }
       else
@@ -309,6 +317,14 @@ int checkOptions(int argc, char **argv, sAudit *_psAudit)
     case 'R':
       _psAudit->iRetries = atoi(optarg);
       break;
+    case 'Z':
+      if (_psAudit->HostType == L_SINGLE)
+      {
+        writeError(ERR_ALERT, "You cannot resume a single host scan.");
+        ret = EXIT_FAILURE;
+      }
+      _psAudit->iResume = atoi(optarg);
+      break;
     default:
       writeError(ERR_CRITICAL, "Unknown error processing command-line options.");
       ret = EXIT_FAILURE;
@@ -664,7 +680,8 @@ int processComboFile(sAudit **_psAudit)
   if (*pComboTmp == '\0')
   {             /* no password specified */
     writeError(ERR_DEBUG, "[processComboFile] No password combo field specified.");
-    if (((*_psAudit)->PassType != L_SINGLE) && ((*_psAudit)->PassType != L_FILE))
+    if (((*_psAudit)->PassType != L_SINGLE) && ((*_psAudit)->PassType != L_FILE) && 
+        ((*_psAudit)->iPasswordBlankFlag == FALSE) && ((*_psAudit)->iPasswordUsernameFlag == FALSE))
     {
       writeError(ERR_FATAL, "Combo format used requires password information via (-p/-P).");
     }
@@ -999,14 +1016,16 @@ int setPassResult(sLogin *_psLogin, char *_pPass)
   pthread_mutex_lock(&_psLogin->psUser->ptmMutex);
 
   writeVerbose(VB_CHECK,
-               "[%s] Host: %s (%d/%d) User: %s (%d/%d) Password: %s (%d/%d)",
+               "[%s] Host: %s (%d of %d, %d complete) User: %s (%d of %d, %d complete) Password: %s (%d of %d complete)",
                _psLogin->psServer->psAudit->pModuleName,
-               _psLogin->psServer->pHostIP,
-               _psLogin->psServer->psAudit->iHostsDone + 1,
+               _psLogin->psServer->psHost->pHost,
+               _psLogin->psServer->psHost->iId,
                _psLogin->psServer->psAudit->iHostCnt,
+               _psLogin->psServer->psAudit->iHostsDone + 1,
                _psLogin->psUser->pUser,
                _psLogin->psUser->iId,
                _psLogin->psServer->psHost->iUserCnt,
+               _psLogin->psServer->psHost->iUsersDone + 1,
                _pPass,
                _psLogin->psUser->iLoginsDone + 1,
                _psLogin->psUser->iPassCnt
@@ -1020,12 +1039,12 @@ int setPassResult(sLogin *_psLogin, char *_pPass)
   {
   case LOGIN_RESULT_SUCCESS:
     if (_psLogin->pErrorMsg) {
-      writeVerbose(VB_FOUND, "[%s] Host: %s User: %s Password: %s [SUCCESS (%s)]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->pHostIP, _psLogin->psUser->pUser, _pPass, _psLogin->pErrorMsg);
+      writeVerbose(VB_FOUND, "[%s] Host: %s User: %s Password: %s [SUCCESS (%s)]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->psHost->pHost, _psLogin->psUser->pUser, _pPass, _psLogin->pErrorMsg);
       free(_psLogin->pErrorMsg);
       _psLogin->pErrorMsg = NULL;
     }
     else
-      writeVerbose(VB_FOUND, "[%s] Host: %s User: %s Password: %s [SUCCESS]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->pHostIP, _psLogin->psUser->pUser, _pPass);
+      writeVerbose(VB_FOUND, "[%s] Host: %s User: %s Password: %s [SUCCESS]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->psHost->pHost, _psLogin->psUser->pUser, _pPass);
     
     _psLogin->psServer->psAudit->iValidPairFound = TRUE;
     _psLogin->psServer->iValidPairFound = TRUE;
@@ -1033,27 +1052,27 @@ int setPassResult(sLogin *_psLogin, char *_pPass)
     break;
   case LOGIN_RESULT_FAIL:
     if (_psLogin->pErrorMsg) {
-      writeError(ERR_INFO, "[%s] Host: %s User: %s [FAILED (%s)]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->pHostIP, _psLogin->psUser->pUser, _psLogin->pErrorMsg);
+      writeError(ERR_INFO, "[%s] Host: %s User: %s [FAILED (%s)]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->psHost->pHost, _psLogin->psUser->pUser, _psLogin->pErrorMsg);
       free(_psLogin->pErrorMsg);
       _psLogin->pErrorMsg = NULL;
     }
     else
-      writeError(ERR_INFO, "[%s] Host: %s User: %s [FAILED]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->pHostIP, _psLogin->psUser->pUser);
+      writeError(ERR_INFO, "[%s] Host: %s User: %s [FAILED]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->psHost->pHost, _psLogin->psUser->pUser);
     
     break;
   case LOGIN_RESULT_ERROR:
     if (_psLogin->pErrorMsg) {
-      writeVerbose(VB_FOUND, "[%s] Host: %s User: %s Password: %s [ERROR (%s)]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->pHostIP, _psLogin->psUser->pUser, _pPass, _psLogin->pErrorMsg);
+      writeVerbose(VB_FOUND, "[%s] Host: %s User: %s Password: %s [ERROR (%s)]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->psHost->pHost, _psLogin->psUser->pUser, _pPass, _psLogin->pErrorMsg);
       free(_psLogin->pErrorMsg);
       _psLogin->pErrorMsg = NULL;
     }
     else
-      writeVerbose(VB_FOUND, "[%s] Host: %s User: %s Password: %s [ERROR]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->pHostIP, _psLogin->psUser->pUser, _pPass);
+      writeVerbose(VB_FOUND, "[%s] Host: %s User: %s Password: %s [ERROR]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->psHost->pHost, _psLogin->psUser->pUser, _pPass);
     
     _psLogin->psUser->iPassStatus = PL_DONE;
     break;
   default:
-    writeError(ERR_INFO, "[%s] Host: %s User: %s [UNKNOWN %d]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->pHostIP, _psLogin->psUser->pUser, _psLogin->iResult);
+    writeError(ERR_INFO, "[%s] Host: %s User: %s [UNKNOWN %d]", _psLogin->psServer->psAudit->pModuleName, _psLogin->psServer->psHost->pHost, _psLogin->psUser->pUser, _psLogin->iResult);
     break;
   }
 
@@ -1119,6 +1138,7 @@ int loadLoginInfo(sAudit *_psAudit)
       psHost->iRetries = _psAudit->iRetries;
       psHost->iUserCnt = 0;
       psHost->iHostStatus = HL_NULL; 
+      psHost->iId = _psAudit->iHostCnt; 
     }
 
     while ((pUser = findNextUser(_psAudit, pUser)))
@@ -1418,6 +1438,7 @@ int startAudit(sAudit *_psAudit)
   void *thread_result;
   int iServerId = 0;
   int iHostDone = 0;
+  int resumeHost = 0;
   int res;
   int ret = SUCCESS;
   sHost *psHost;
@@ -1447,11 +1468,48 @@ int startAudit(sAudit *_psAudit)
   memset(pptServer, 0, _psAudit->iServerCnt * sizeof(pptServer));
 
   psHost = _psAudit->psHostRoot;
+
+  for (; _psAudit->iResume > 1; _psAudit->iResume--)
+  {
+    psHost = psHost->psHostNext;
+    iHostDone = iHostDone + 1;
+  }
+
   while (iHostDone < _psAudit->iHostCnt)
   {
+    if (ctrlc == 1)
+    {
+      for (iServerId = 0; iServerId < _psAudit->iServerCnt; iServerId++)
+      {
+        resumeHost = psServer[0].psHost->iId;
+        //Following can be usefule for debugging
+        /* printf(
+        "[%s] Host: %s (%d of %d, %d complete)\n",
+              psServer[iServerId].psAudit->pModuleName,
+              psServer[iServerId].pHostIP,
+              psServer[iServerId].psHost->iId,
+              psServer[iServerId].psAudit->iHostCnt,
+              psServer[iServerId].psAudit->iHostsDone
+        // may use some later for adding in user/passwd resume
+              //_psLogin->psUser->pUser,
+              //_psLogin->psUser->iId,
+              //psServer[iServerId]->psHost->iUserCnt,
+              //psServer[iServerId]->psHost->iUsersDone + 1,
+              //_pPass,
+              //_psLogin->psUser->iLoginsDone + 1,
+              //_psLogin->psUser->iPassCnt
+             ); */
+
+        if (psServer[iServerId].psHost->iId < resumeHost)
+          resumeHost = psServer[iServerId].psHost->iId;
+      }
+
+      writeError(VB_IMPORTANT,"SIGINT Caught - Resume scan from: %d",resumeHost);
+      exit(1);
+    }
+
     for (iServerId = 0; iServerId < _psAudit->iServerCnt; iServerId++)
     {
-
       if ((psHost) && (psServer[iServerId].iStatus == SRV_IDLE))
       {
         writeError(ERR_DEBUG_AUDIT, "[audit] starting new server: %d", iServerId);
@@ -1501,6 +1559,21 @@ int startAudit(sAudit *_psAudit)
   return ret;
 }
 
+void sigint_handler(int sig)
+{
+  char answer[1];
+
+  ctrlc = 1;
+  signal(SIGINT, sigint_handler);
+  //writeVerbose(VB_GENERAL, "At current host: %s", );
+  //printf("Do you wish to continue or quit?");
+  //scanf("%s",answer);
+  //if (*answer == 'c')
+  //  signal(SIGINT, sigint_handler);
+  //else
+  //  exit(1);
+}
+
 int main(int argc, char **argv, char *envp[])
 {
   int iRes, iExitStatus;
@@ -1513,6 +1586,8 @@ int main(int argc, char **argv, char *envp[])
 
   szModuleName = NULL;
 
+  // Added signal handling - for SIGINT
+  signal(SIGINT, sigint_handler);
   // Don't worry if there are NULL or blank values here - they will be checked when loading the module
   szModulePaths[0] = getenv("MEDUSA_MODULE_PATH");
   szModulePaths[1] = ".";
