@@ -36,19 +36,27 @@
  ***************************************************************************/
 
 #include "module.h"
+#include <math.h>
 
 #define	MODULE_NAME		"http.mod"
 #define MODULE_AUTHOR  "fizzgig <fizzgig@foofus.net>"
 #define	MODULE_SUMMARY_USAGE	"Brute force module for HTTP"
 #define MODULE_VERSION		"1.3.0"
-#define MODULE_VERSION_SVN "$Id: http.c 812 2007-08-28 21:20:17Z jmk $"
+#define MODULE_VERSION_SVN "$Id: http.c 980 2008-10-01 20:22:51Z jmk $"
 #define MODULE_SUMMARY_FORMAT	"%s : version %s"
 #define MODULE_SUMMARY_FORMAT_WARN  "%s : version %s (%s)"
 #define OPENSSL_WARNING "No usable OPENSSL. Module disabled."
 
+#define FREE(x) \
+        if (x != NULL) { \
+           free(x); \
+           x = NULL; \
+        }
+
 #ifdef HAVE_LIBSSL
 
-#include "../http-ntlm.h"
+#include "ntlm.h"
+#include "http-digest.h"
 
 #define HTTP_PORT 80
 #define HTTPS_PORT 443
@@ -57,6 +65,7 @@
 #define AUTH_NONE 1
 #define AUTH_BASIC 2
 #define AUTH_NTLM 3
+#define AUTH_DIGEST 4
 
 typedef struct __MODULE_DATA {
   //char *szDomain;
@@ -180,7 +189,7 @@ int go(sLogin* logins, int argc, char *argv[])
         writeError(ERR_WARNING, "Invalid method: %s.", pOpt);
       }
 
-      free(pOptTmp);
+      FREE(pOptTmp);
     }
 
     initHttp(psSessionData, logins);
@@ -339,11 +348,11 @@ int getAuthType(int hSocket, _MODULE_DATA* _psSessionData, sLogin** login, int n
   if (medusaSend(hSocket, bufSend, nSendBufferSize, 0) < 0)
   {
     writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
-    free(bufSend);
+    FREE(bufSend);
     return FAILURE;
   }
   
-  free(bufSend);
+  FREE(bufSend);
 
   bufReceive = medusaReceiveLine(hSocket, &nReceiveBufferSize);
   if (bufReceive == NULL)
@@ -352,12 +361,6 @@ int getAuthType(int hSocket, _MODULE_DATA* _psSessionData, sLogin** login, int n
     return FAILURE;
   }
   
-  while (((char *)strcasestr(bufReceive, "WWW-Authenticate") == NULL) && (bufReceive[0] != '\0'))
-  {
-    free(bufReceive);
-    bufReceive = medusaReceiveLine(hSocket, &nReceiveBufferSize);
-  }
-
   writeError(ERR_DEBUG_MODULE, "[%s] Parsing authentication header: %s", MODULE_NAME, bufReceive);
   if (strcasestr(bufReceive, "WWW-Authenticate: Basic"))
   {
@@ -369,6 +372,11 @@ int getAuthType(int hSocket, _MODULE_DATA* _psSessionData, sLogin** login, int n
     writeError(ERR_DEBUG_MODULE, "[%s] Server requested integrated windows authentication.", MODULE_NAME);
     _psSessionData->nAuthType = AUTH_NTLM;
   }
+  else if ((strcasestr(bufReceive, "WWW-Authenticate: Digest")) || strcasestr(bufReceive, "WWW-Authenticate:Digest"))
+  {
+    writeError(ERR_DEBUG_MODULE, "[%s] Server requested digest authentication.", MODULE_NAME);
+    _psSessionData->nAuthType = AUTH_DIGEST;
+  }
   else if (strcasestr(bufReceive, "WWW-Authenticate:"))
   {
     writeError(ERR_ERROR, "[%s] Server requested unknown authentication type.", MODULE_NAME);
@@ -380,7 +388,7 @@ int getAuthType(int hSocket, _MODULE_DATA* _psSessionData, sLogin** login, int n
     _psSessionData->nAuthType = AUTH_NONE;
   }
 
-  free(bufReceive);
+  FREE(bufReceive);
   return(SUCCESS);
 }
 
@@ -411,8 +419,8 @@ int sendAuthBasic(int hSocket, _MODULE_DATA* _psSessionData, char* szLogin, char
     nRet = FAILURE;  
   }
   
-  free(szEncodedAuth);
-  free(bufSend);
+  FREE(szEncodedAuth);
+  FREE(bufSend);
   return nRet;
 }
 
@@ -422,30 +430,35 @@ int sendAuthNTLM(int hSocket, _MODULE_DATA* _psSessionData, char* szLogin, char*
   unsigned char* bufReceive;
   int nReceiveBufferSize = 0;
   int nSendBufferSize = 0;
-  int i, nState;
-  struct ntlmdata ntlm;
-  unsigned char* szEncodedAuth, *szTmp;
+  unsigned char *szTmpBuf;
+  unsigned char *szTmpBuf64;
+  unsigned char *szTmpChall;
 
-  nState = 0;
-  memset(&ntlm, 0, sizeof(ntlm));
-  
-  /* Send NTLM request */
-  szEncodedAuth = ntlm_output(&ntlm, szLogin, szPassword, &nState);
-  if (szEncodedAuth == NULL)
+  /* --- Send Type-1 NTLM request --- */
+
+  szTmpBuf = malloc( sizeof(tSmbNtlmAuthRequest) );
+  memset(szTmpBuf, 0, sizeof(tSmbNtlmAuthRequest));  
+  szTmpBuf64 = malloc(2 * sizeof(tSmbNtlmAuthRequest) + 2);
+  memset(szTmpBuf64, 0, 2 * sizeof(tSmbNtlmAuthRequest) + 2);
+
+  buildAuthRequest((tSmbNtlmAuthRequest*)szTmpBuf, 0x00000202, NULL, NULL);
+  if (szTmpBuf == NULL)
   {
     writeError(ERR_ERROR, "[%s] Failed to create initial NTLM request.", MODULE_NAME);
     return FAILURE;
   }
-  writeError(ERR_DEBUG_MODULE, "[%s] Sending initial challenge (B64 Encoded): %s", MODULE_NAME, szEncodedAuth);
+
+  base64_encode(szTmpBuf, sizeof(tSmbNtlmAuthRequest), szTmpBuf64);
+  writeError(ERR_DEBUG_MODULE, "[%s] Sending initial challenge (B64 Encoded): %s", MODULE_NAME, szTmpBuf64);
   
   nSendBufferSize = 5 + strlen(_psSessionData->szDir) + 17 + strlen(_psSessionData->szHostHeader) +
-                    14 + strlen(_psSessionData->szUserAgent) + 17 + strlen(szEncodedAuth) + 28;
+                    14 + strlen(_psSessionData->szUserAgent) + 22 + strlen(szTmpBuf64) + 28;
  
   bufSend = malloc(nSendBufferSize + 1);
   memset(bufSend, 0, nSendBufferSize + 1);
 
-  sprintf(bufSend, "GET /%s HTTP/1.1\r\nHost: %s\r\nUser-Agent: %s\r\nAuthorization: %s\r\nConnection: close\r\n\r\n", 
-          _psSessionData->szDir, _psSessionData->szHostHeader, _psSessionData->szUserAgent, szEncodedAuth);
+  sprintf(bufSend, "GET /%s HTTP/1.1\r\nHost: %s\r\nUser-Agent: %s\r\nAuthorization: NTLM %s\r\nConnection: keep-alive\r\n\r\n", 
+          _psSessionData->szDir, _psSessionData->szHostHeader, _psSessionData->szUserAgent, szTmpBuf64);
   
   if (medusaSend(hSocket, bufSend, nSendBufferSize, 0) < 0)
   {
@@ -453,9 +466,9 @@ int sendAuthNTLM(int hSocket, _MODULE_DATA* _psSessionData, char* szLogin, char*
     return FAILURE;
   }
   
-  free(szEncodedAuth);
-  free(bufSend);
-
+  FREE(szTmpBuf);
+  FREE(szTmpBuf64);
+  FREE(bufSend);
 
   /* Retrieve NTLM challenge from server */
   bufReceive = medusaReceiveLine(hSocket, &nReceiveBufferSize);
@@ -465,52 +478,58 @@ int sendAuthNTLM(int hSocket, _MODULE_DATA* _psSessionData, char* szLogin, char*
     return FAILURE;
   }
   
-  while (((char *)strcasestr(bufReceive, "WWW-Authenticate: NTLM ") == NULL) && (bufReceive[0] != '\0'))
-  {
-    free(bufReceive);
-    bufReceive = medusaReceiveLine(hSocket, &nReceiveBufferSize);
-  }
-
   if (bufReceive[0] == '\0')
   {
-    writeError(ERR_ERROR, "[%s] Failed to locate NTLM challenge.", MODULE_NAME);
+    writeError(ERR_ERROR, "[%s] Service did not respond to our authentication request.", MODULE_NAME);
     return FAILURE;
   }
 
-  /* Extract NTLM challenge from response */
-  szEncodedAuth = (char *)strcasestr(bufReceive, "WWW-Authenticate: NTLM ");
-  szEncodedAuth += 18;
-  szTmp = ((char*)index(szEncodedAuth, '\r'));
-  szTmp[0] = '\0';
-  writeError(ERR_DEBUG_MODULE, "[%s] NTLM Challenge (B64 Encoded): %s", MODULE_NAME, szEncodedAuth);
-  
-  if (ntlm_input(&ntlm, szEncodedAuth) == 0)
+  /* --- Extract NTLM challenge from Type-2 NTLM response --- */
+
+  szTmpBuf64 = (char *)strcasestr(bufReceive, "WWW-Authenticate: NTLM ");
+  if (szTmpBuf64 == NULL)
   {
-    writeError(ERR_ERROR, "[%s] Failed to process NTLM response.", MODULE_NAME);
+    writeError(ERR_ERROR, "[%s] Failed to locate NTLM challenge within server response.", MODULE_NAME);
     return FAILURE;
   }
- 
-  free(bufReceive);
+
+  szTmpBuf = ((char*)index(szTmpBuf64, '\r'));
+  szTmpBuf[0] = '\0';
+
+  szTmpChall = malloc( sizeof(tSmbNtlmAuthChallenge) );
+  memset(szTmpChall, 0, sizeof(tSmbNtlmAuthChallenge));
+
+  writeError(ERR_DEBUG_MODULE, "[%s] NTLM Challenge (B64 Encoded): %s", MODULE_NAME, szTmpBuf64 + 23);
+  base64_decode(szTmpBuf64 + 23, szTmpChall);
   
+  FREE(bufReceive);
  
-  /* Send NTLM response */
-  nState = 1;
-  szEncodedAuth = ntlm_output(&ntlm, szLogin, szPassword, &nState);
-  if (szEncodedAuth == NULL)
+  /* --- Send Type-3 NTLM reply --- */
+
+  szTmpBuf = malloc( sizeof(tSmbNtlmAuthResponse) );
+  memset(szTmpBuf, 0, sizeof(tSmbNtlmAuthResponse));
+
+  szTmpBuf64 = malloc( 2 * sizeof(tSmbNtlmAuthResponse) + 2);
+  memset(szTmpBuf64, 0, 2 * sizeof(tSmbNtlmAuthResponse) + 2);
+
+  buildAuthResponse((tSmbNtlmAuthChallenge*)szTmpChall, (tSmbNtlmAuthResponse*)szTmpBuf, 0, szLogin, szPassword, NULL, NULL);
+  if (szTmpBuf == NULL)
   {
     writeError(ERR_ERROR, "[%s] Failed to create NTLM response.", MODULE_NAME);
     return FAILURE;
   }
-  writeError(ERR_DEBUG_MODULE, "[%s] NTLM Response (B64 Encoded): %s", MODULE_NAME, szEncodedAuth);
+  
+  base64_encode(szTmpBuf, sizeof(tSmbNtlmAuthResponse), szTmpBuf64);
+  writeError(ERR_DEBUG_MODULE, "[%s] NTLM Response (B64 Encoded): %s", MODULE_NAME, szTmpBuf64);
   
   nSendBufferSize = 5 + strlen(_psSessionData->szDir) + 17 + strlen(_psSessionData->szHostHeader) +
-                    14 + strlen(_psSessionData->szUserAgent) + 17 + strlen(szEncodedAuth) + 28;
+                    14 + strlen(_psSessionData->szUserAgent) + 22 + strlen(szTmpBuf64) + 23;
  
   bufSend = malloc(nSendBufferSize + 1);
   memset(bufSend, 0, nSendBufferSize + 1);
 
-  sprintf(bufSend, "GET /%s HTTP/1.1\r\nHost: %s\r\nUser-Agent: %s\r\nAuthorization: %s\r\nConnection: close\r\n\r\n", 
-          _psSessionData->szDir, _psSessionData->szHostHeader, _psSessionData->szUserAgent, szEncodedAuth);
+  sprintf(bufSend, "GET /%s HTTP/1.1\r\nHost: %s\r\nUser-Agent: %s\r\nAuthorization: NTLM %s\r\nConnection: close\r\n\r\n", 
+          _psSessionData->szDir, _psSessionData->szHostHeader, _psSessionData->szUserAgent, szTmpBuf64);
   
   if (medusaSend(hSocket, bufSend, nSendBufferSize, 0) < 0)
   {
@@ -518,12 +537,289 @@ int sendAuthNTLM(int hSocket, _MODULE_DATA* _psSessionData, char* szLogin, char*
     return FAILURE;
   }
 
-  free(szEncodedAuth);
-  free(bufSend);
+  FREE(szTmpBuf);
+  FREE(szTmpBuf64);
+  FREE(bufSend);
 
   return SUCCESS;
 }
 
+/* http://www.ietf.org/rfc/rfc2617.txt */
+int sendAuthDigest(int hSocket, _MODULE_DATA* _psSessionData, char* szLogin, char* szPassword)
+{
+  unsigned char* bufSend;
+  unsigned char* bufReceive;
+  int nReceiveBufferSize = 0;
+  int nSendBufferSize = 0;
+  int i;
+  unsigned char *szTmp = NULL;
+  unsigned char *szTmp1 = NULL;
+  unsigned char *szAuthenticate = NULL;
+  unsigned char *szAuthorization = NULL;
+
+  char *szNonce = NULL;
+  char *szCNonce = "31337";
+  char *szRealm = NULL; 
+  char *szAlg = NULL;
+  char  szNonceCount[9] = "00000001";
+  char *szMethod = "GET";
+  char *szQop = NULL;
+  char *szURI = _psSessionData->szDir;
+  char *szOpaque = NULL;
+  HASHHEX HA1;
+  HASHHEX HA2 = "";
+  HASHHEX Response;
+    
+  /* Send initial request */
+  writeError(ERR_DEBUG_MODULE, "[%s] Sending initial request for digest authentication.", MODULE_NAME);
+  
+  nSendBufferSize = 5 + strlen(_psSessionData->szDir) + 17 + strlen(_psSessionData->szHostHeader) +
+                    14 + strlen(_psSessionData->szUserAgent) + 28;
+
+  bufSend = malloc(nSendBufferSize + 1);
+  memset(bufSend, 0, nSendBufferSize + 1);
+
+  sprintf(bufSend, "GET /%s HTTP/1.1\r\nHost: %s\r\nUser-Agent: %s\r\nConnection: keep-alive\r\n\r\n", 
+          _psSessionData->szDir, _psSessionData->szHostHeader, _psSessionData->szUserAgent);
+  
+  if (medusaSend(hSocket, bufSend, nSendBufferSize, 0) < 0)
+  {
+    writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
+    return FAILURE;
+  }
+  
+  FREE(bufSend);
+
+  /* Retrieve digest challenge from server */
+  bufReceive = medusaReceiveLine(hSocket, &nReceiveBufferSize);
+  if (bufReceive == NULL)
+  {
+    writeError(ERR_ERROR, "[%s] No data received", MODULE_NAME);
+    return FAILURE;
+  }
+  
+  if (bufReceive[0] == '\0')
+  {
+    writeError(ERR_ERROR, "[%s] Failed to locate digest challenge.", MODULE_NAME);
+    return FAILURE;
+  }
+
+  /* Parse WWW-Authenticate Digest Response */
+  /* Example: WWW-Authenticate: Digest realm="Inter-Tel 5000 (00103605AB8A)", nonce="86591bebf1330b5e57b8de2e4ac216b2", qop="auth" */
+  if ( szTmp = (char *)strcasestr(bufReceive, "WWW-Authenticate: Digest ") )
+  {
+    szTmp += 18;
+  }
+  else if ( szTmp = (char *)strcasestr(bufReceive, "WWW-Authenticate:Digest ") )
+  {
+    szTmp += 17;
+  }
+  else
+  {
+    writeError(ERR_ERROR, "[%s] Failed to locate digest challenge.", MODULE_NAME);
+    return FAILURE;
+  }
+  szTmp1 = ((char*)index(szTmp, '\r'));
+
+  szAuthenticate = malloc (szTmp1 - szTmp + 1);
+  memset(szAuthenticate, 0, szTmp1 - szTmp + 1);
+  strncpy(szAuthenticate, szTmp, szTmp1 - szTmp);
+  
+  FREE(bufReceive);
+
+  writeError(ERR_DEBUG_MODULE, "[%s] Server WWW-Authenticate Digest Response: %s", MODULE_NAME, szAuthenticate);
+
+  /* Extract Digest Algorithm, if Specified */
+  /* We currently only support MD5 and MD5-Sess (session) - Do others exist? */
+  if ( strcasestr(szAuthenticate, "algorithm=MD5-sess") || strcasestr(szAuthenticate, "algorithm=\"MD5-sess\"") )
+  {
+    writeError(ERR_DEBUG_MODULE, "[%s] Server requested Digest MD5-sess algorithm.", MODULE_NAME);
+    szAlg = malloc(9);
+    memset(szAlg, 0, 9);
+    sprintf(szAlg, "MD5-sess"); 
+  }
+  else if ( strcasestr(szAuthenticate, "algorithm=MD5") || strcasestr(szAuthenticate, "algorithm=\"MD5\""))
+  {
+    writeError(ERR_DEBUG_MODULE, "[%s] Server requested Digest MD5 algorithm.", MODULE_NAME);
+    szAlg = malloc(4);
+    memset(szAlg, 0, 4);
+    sprintf(szAlg, "MD5"); 
+  }
+  else if ( strcasestr(szAuthenticate, "algorithm=") )
+  {
+    writeError(ERR_ERROR, "[%s] Server requested unknown Digest algorithm.", MODULE_NAME);
+    return FAILURE;
+  }
+  else
+  {
+    writeError(ERR_DEBUG_MODULE, "[%s] Server did not specify a Digest algorithm, so we're assuming MD5.", MODULE_NAME);
+    szAlg = malloc(4);
+    memset(szAlg, 0, 4);
+    sprintf(szAlg, "MD5"); 
+  }
+
+  /* Extract Digest Realm */
+  szTmp = (char *)strcasestr(szAuthenticate, "realm=\"");
+  if (szTmp)
+  {
+    szTmp += 7;
+    szTmp1 = ((char*)index(szTmp, '"'));
+
+    szRealm = malloc (szTmp1 - szTmp + 1);
+    memset(szRealm, 0, szTmp1 - szTmp + 1);
+    strncpy(szRealm, szTmp, szTmp1 - szTmp);
+    writeError(ERR_DEBUG_MODULE, "[%s] Extracted Realm Response: %s", MODULE_NAME, szRealm);
+  }
+  else
+  {
+    writeError(ERR_ERROR, "[%s] Failed to extract server Realm response.", MODULE_NAME);
+    szRealm = malloc(1);
+    memset(szRealm, 0, 1);
+  }
+
+  /* Extract Digest Server Nonce */
+  szTmp = (char *)strcasestr(szAuthenticate, "nonce=\"");
+  if (szTmp)
+  {
+    szTmp += 7;
+    szTmp1 = ((char*)index(szTmp, '"'));
+
+    szNonce = malloc (szTmp1 - szTmp + 1);
+    memset(szNonce, 0, szTmp1 - szTmp + 1);
+    strncpy(szNonce, szTmp, szTmp1 - szTmp);
+    writeError(ERR_DEBUG_MODULE, "[%s] Extracted Nonce Response: %s", MODULE_NAME, szNonce);
+  }
+  else
+  {
+    writeError(ERR_ERROR, "[%s] Failed to extract server Nonce response.", MODULE_NAME);
+    szNonce = malloc(1);
+    memset(szNonce, 0, 1);
+  }
+  
+  /* Extract Digest Quality of Protection (QoP) - If Specified */
+  szTmp = (char *)strcasestr(szAuthenticate, "qop=\"");
+  if (szTmp)
+  {
+    szTmp += 5;
+    szTmp1 = ((char*)index(szTmp, '"'));
+
+    szQop = malloc (szTmp1 - szTmp + 1);
+    memset(szQop, 0, szTmp1 - szTmp + 1);
+    strncpy(szQop, szTmp, szTmp1 - szTmp);
+    writeError(ERR_DEBUG_MODULE, "[%s] Extracted Quality of Protection (QoP) Response: %s", MODULE_NAME, szQop);
+  }
+  else
+  {
+    writeError(ERR_DEBUG_MODULE, "[%s] Failed to extract server Quality of Protection (QoP) response.", MODULE_NAME);
+    szQop = NULL;
+  }
+  
+  /* Extract Digest Opaque Value - If Specified */
+  szTmp = (char *)strcasestr(szAuthenticate, "opaque=\"");
+  if (szTmp)
+  {
+    szTmp += 8;
+    szTmp1 = ((char*)index(szTmp, '"'));
+
+    szOpaque = malloc (szTmp1 - szTmp + 1);
+    memset(szOpaque, 0, szTmp1 - szTmp + 1);
+    strncpy(szOpaque, szTmp, szTmp1 - szTmp);
+    writeError(ERR_DEBUG_MODULE, "[%s] Extracted Server Opaque Value: %s", MODULE_NAME, szOpaque);
+  }
+  else
+  {
+    writeError(ERR_DEBUG_MODULE, "[%s] Failed to extract server Opaque value.", MODULE_NAME);
+    szOpaque = NULL;
+  }
+  
+  FREE(szAuthenticate);
+ 
+  /* Send digest response */
+  /* Example: Authorization: Digest username="it5k", realm="Inter-Tel 5000 (00103605AB8A)", nonce="94144a2abae7411d0f6af2b533425497", uri="/", 
+                             response="b9c3980ae4e7fb69796772a3bacc5c18"\r\n */
+ 
+  writeError(ERR_DEBUG_MODULE, "[%s] szAlg: %s", MODULE_NAME, szAlg);
+  writeError(ERR_DEBUG_MODULE, "[%s] szLogin: %s", MODULE_NAME, szLogin);
+  writeError(ERR_DEBUG_MODULE, "[%s] szRealm: %s", MODULE_NAME, szRealm);
+  writeError(ERR_DEBUG_MODULE, "[%s] szPassword: %s", MODULE_NAME, szPassword);
+  writeError(ERR_DEBUG_MODULE, "[%s] szNonce: %s", MODULE_NAME, szNonce);
+  writeError(ERR_DEBUG_MODULE, "[%s] szCNonce: %s", MODULE_NAME, szCNonce);
+  writeError(ERR_DEBUG_MODULE, "[%s] szNonceCount: %s", MODULE_NAME, szNonceCount);
+  writeError(ERR_DEBUG_MODULE, "[%s] szQop: %s", MODULE_NAME, szQop);
+  writeError(ERR_DEBUG_MODULE, "[%s] szOpaque: %s", MODULE_NAME, szOpaque);
+  writeError(ERR_DEBUG_MODULE, "[%s] szMethod: %s", MODULE_NAME, szMethod);
+  writeError(ERR_DEBUG_MODULE, "[%s] szURI: %s", MODULE_NAME, szURI);
+
+  DigestCalcHA1(szAlg, szLogin, szRealm, szPassword, szNonce, szCNonce, HA1);
+  DigestCalcResponse(HA1, szNonce, szNonceCount, szCNonce, szQop, szMethod, szURI, HA2, Response);
+  writeError(ERR_DEBUG_MODULE, "[%s] Calculated Digest Response: %s", MODULE_NAME, Response);
+  
+  /*
+    BASE:   Digest username=\"%s\", realm=\"%s\", nonce=\"%s\", uri=\"%s\", algorithm=\"%s\", response=\"%s\"
+    QOP:    , qop=%s, nc=%s, cnonce=\"%s\"
+    OPAQUE: , opaque=\"%s\"
+  */
+
+  nSendBufferSize = 17 + strlen(szLogin) + 10 + strlen(szRealm) + 10 + strlen(szNonce) + 8 + strlen(szURI) + 14 + strlen(szAlg) + 13 + strlen(Response);
+
+  /* If the server specified a QoP, the client must use a cnonce */
+  if ( strcasestr(szQop, "auth-int") )
+  {
+    writeError(ERR_ERROR, "[%s] Integrity protection (i.e. qop: auth-int) is currently not supported.", MODULE_NAME);
+    return FAILURE;
+  }
+  else if ( strcasestr(szQop, "auth") )
+  {
+    nSendBufferSize += 7 + strlen(szQop) + 5 + strlen(szNonceCount) + 10 + strlen(szCNonce) + 1;
+  }
+
+  /* If the server specified an opaque value, that same value should be included in our response. */
+  if ( szOpaque )
+  {
+    nSendBufferSize += 10 + strlen(szOpaque) + 1; 
+  }
+
+  szAuthorization = malloc(nSendBufferSize + 1);
+  memset(szAuthorization, 0, nSendBufferSize + 1);
+ 
+  if ( (szQop != NULL) && (szOpaque != NULL) )
+    sprintf(szAuthorization, "Digest username=\"%s\", realm=\"%s\", nonce=\"%s\", uri=\"%s\", algorithm=%s, response=\"%s\", qop=%s, nc=00000001, cnonce=\"%s\", opaque=\"%s\"",
+                             szLogin, szRealm, szNonce, szURI, szAlg, Response, szQop, szCNonce, szOpaque);
+  else if (szQop != NULL)
+    sprintf(szAuthorization, "Digest username=\"%s\", realm=\"%s\", nonce=\"%s\", uri=\"%s\", algorithm=%s, response=\"%s\", qop=%s, nc=00000001, cnonce=\"%s\"",
+                             szLogin, szRealm, szNonce, szURI, szAlg, Response, szQop, szCNonce);
+  else if (szOpaque != NULL)
+    sprintf(szAuthorization, "Digest username=\"%s\", realm=\"%s\", nonce=\"%s\", uri=\"%s\", algorithm=%s, response=\"%s\", opaque=\"%s\"",
+                             szLogin, szRealm, szNonce, szURI, szAlg, Response, szOpaque);
+  else
+    sprintf(szAuthorization, "Digest username=\"%s\", realm=\"%s\", nonce=\"%s\", uri=\"%s\", algorithm=%s, response=\"%s\"",
+                             szLogin, szRealm, szNonce, szURI, szAlg, Response);
+
+  FREE(szAlg);
+  FREE(szRealm);
+  FREE(szNonce);
+  FREE(szQop);
+  FREE(szOpaque);
+
+  nSendBufferSize = 5 + strlen(_psSessionData->szDir) + 17 + strlen(_psSessionData->szHostHeader) +
+                    14 + strlen(_psSessionData->szUserAgent) + 17 + strlen(szAuthorization) + 28;
+
+  bufSend = malloc(nSendBufferSize + 1);
+  memset(bufSend, 0, nSendBufferSize + 1);
+
+  sprintf(bufSend, "GET /%s HTTP/1.1\r\nHost: %s\r\nUser-Agent: %s\r\nAuthorization: %s\r\nConnection: keep-alive\r\n\r\n", 
+          _psSessionData->szDir, _psSessionData->szHostHeader, _psSessionData->szUserAgent, szAuthorization);
+  
+  if (medusaSend(hSocket, bufSend, nSendBufferSize, 0) < 0)
+  {
+    writeError(ERR_ERROR, "[%s] failed: medusaSend was not successful", MODULE_NAME);
+    return FAILURE;
+  }
+
+  FREE(szAuthorization);
+  FREE(bufSend);
+}
+
 int tryLogin(int hSocket, _MODULE_DATA* _psSessionData, sLogin** login, int nPort, char* szLogin, char* szPassword)
 {
   char* pReceiveBuffer;
@@ -546,6 +842,10 @@ int tryLogin(int hSocket, _MODULE_DATA* _psSessionData, sLogin** login, int nPor
       writeError(ERR_DEBUG_MODULE, "[%s] Sending Windows Integrated (NTLM) Authentication.", MODULE_NAME);
       nRet = sendAuthNTLM(hSocket, _psSessionData, szLogin, szPassword);
       break;
+    case AUTH_DIGEST:
+      writeError(ERR_DEBUG_MODULE, "[%s] Sending Digest Authentication.", MODULE_NAME);
+      nRet = sendAuthDigest(hSocket, _psSessionData, szLogin, szPassword);
+      break;
     default:
       break;
   }
@@ -570,7 +870,7 @@ int tryLogin(int hSocket, _MODULE_DATA* _psSessionData, sLogin** login, int nPor
     
   while (strstr(pReceiveBuffer, "HTTP/1.") == NULL && (pReceiveBuffer[0] != '\0'))
   {
-    free(pReceiveBuffer);
+    FREE(pReceiveBuffer);
     pReceiveBuffer = medusaReceiveLine(hSocket, &nReceiveBufferSize);
   }
 
