@@ -36,12 +36,13 @@
 #include <string.h>
 #include <stdlib.h>
 #include "module.h"
+#include "ntlm.h"
 
 #define MODULE_NAME    "imap.mod"
 #define MODULE_AUTHOR  "pMonkey <pmonkey@foofus.net>"
 #define MODULE_SUMMARY_USAGE  "Brute force module for IMAP sessions"
-#define MODULE_VERSION    "1.1.0"
-#define MODULE_VERSION_SVN "$Id: imap.c 613 2006-10-31 22:56:11Z jmk $"
+#define MODULE_VERSION    "1.2.0"
+#define MODULE_VERSION_SVN "$Id: imap.c 987 2008-10-10 17:43:41Z jmk $"
 #define MODULE_SUMMARY_FORMAT  "%s : version %s"
 
 #define FREE(x) \
@@ -58,6 +59,7 @@
 #define AUTH_UNKNOWN 0
 #define AUTH_LOGIN 1
 #define AUTH_PLAIN 2
+#define AUTH_NTLM 3
 
 typedef struct __MODULE_DATA {
   char *szTag;
@@ -73,7 +75,7 @@ enum MODULE_STATE
 };
 
 // Forward declarations
-int getAuthType(int hSocket, _MODULE_DATA* _psSessionData);
+int initConnection(_MODULE_DATA *_psSessionData, int hSocket, sConnectParams *params);
 int tryLogin(int hSocket,  _MODULE_DATA* _psSessionData, sLogin** login, char* szLogin, char* szPassword);
 int initModule(_MODULE_DATA* _psSessionData, sLogin* login);
 
@@ -108,7 +110,7 @@ void showUsage()
   writeVerbose(VB_NONE, "%s (%s) %s :: %s\n", MODULE_NAME, MODULE_VERSION, MODULE_AUTHOR, MODULE_SUMMARY_USAGE);
   writeVerbose(VB_NONE, "Available module options:");
   writeVerbose(VB_NONE, "  TAG:? (Default: gerg)");
-  writeVerbose(VB_NONE, "  AUTH:? (Authentication Type (LOGIN/PLAIN). Default: automatic)");
+  writeVerbose(VB_NONE, "  AUTH:? (Authentication Type (LOGIN/PLAIN/NTLM). Default: automatic)");
   writeVerbose(VB_NONE, "");
   writeVerbose(VB_NONE, "Usage example: \"-M imap -m TAG:A0001 -m AUTH:PLAIN");
 }
@@ -164,6 +166,8 @@ int go(sLogin* logins, int argc, char *argv[])
           psSessionData->nAuthType = AUTH_LOGIN;
         else if (strcmp(pOpt, "PLAIN") == 0)
           psSessionData->nAuthType = AUTH_PLAIN;
+        else if (strcmp(pOpt, "NTLM") == 0)
+          psSessionData->nAuthType = AUTH_NTLM;
         else
           writeError(ERR_WARNING, "Invalid value for method AUTH.");
       }
@@ -247,25 +251,15 @@ int initModule(_MODULE_DATA *_psSessionData, sLogin* psLogin)
           return FAILURE;
         }
 
-        /* Retrieve IMAP server banner */
-        bufReceive = medusaReceiveRaw(hSocket, &nReceiveBufferSize);
-        if (bufReceive == NULL)
+        if (initConnection(_psSessionData, hSocket, &params) == FAILURE)
         {
-          writeError(ERR_ERROR, "[%s] Failed to retrieve IMAP server banner. Exiting...", MODULE_NAME);
-          return FAILURE; 
+          psLogin->iResult = LOGIN_RESULT_UNKNOWN;
+          psLogin->iStatus = LOGIN_FAILED;
+          FREE(_psSessionData);
+          return FAILURE;
         }
 
-        if (_psSessionData->nAuthType == AUTH_UNKNOWN)
-        {
-          getAuthType(hSocket, _psSessionData);
-
-          if (_psSessionData->nAuthType == AUTH_UNKNOWN)
-          {
-            psLogin->iResult = LOGIN_RESULT_UNKNOWN;
-            psLogin->iStatus = LOGIN_FAILED;
-            return FAILURE;
-          }
-        }
+        writeError(ERR_DEBUG_MODULE, "Connected");
 
         nState = MSTATE_RUNNING;
         break;
@@ -296,16 +290,40 @@ int initModule(_MODULE_DATA *_psSessionData, sLogin* psLogin)
 }
 
 /* Module Specific Functions */
-
-int getAuthType(int hSocket, _MODULE_DATA* _psSessionData)
+int initConnection(_MODULE_DATA *_psSessionData, int hSocket, sConnectParams *params)
 {
-  int iRet;
-  char* bufReceive;
-  unsigned char* bufSend;
-  int nSendBufferSize = 0;
+  unsigned char *bufSend = NULL;
+  unsigned char *bufReceive = NULL;
   int nReceiveBufferSize = 0;
+  int nSendBufferSize = 0;
 
-  /* Let's ask for capability even though most give it out by default */
+  /* Retrieve IMAP server banner */
+  bufReceive = medusaReceiveRaw(hSocket, &nReceiveBufferSize);
+  if (bufReceive == NULL)
+  {
+    writeError(ERR_ERROR, "[%s] Failed to retrieve IMAP server banner. Exiting...", MODULE_NAME);
+    return FAILURE; 
+  }
+  else if ((strstr(bufReceive,"* OK ") != NULL))
+  {
+    writeError(ERR_DEBUG_MODULE, "[%s] Received IMAP server banner: %s", MODULE_NAME, bufReceive);
+    FREE(bufReceive);
+  }
+  else if ((strstr(bufReceive,"* BYE Connection refused") != NULL))
+  {
+    writeError(ERR_ERROR, "[%s] IMAP server refused connection. Is SSL required?", MODULE_NAME);
+    FREE(bufReceive);
+    return FAILURE;
+  }
+  else
+  {
+    writeError(ERR_ERROR, "[%s] Failed to retrieve IMAP server banner.", MODULE_NAME);
+    FREE(bufReceive);
+    return FAILURE;
+  }
+
+  /* Request IMAP server capabilities */
+  writeError(ERR_DEBUG_MODULE, "[%s] Sending IMAP CAPABILITIES request.", MODULE_NAME);
   nSendBufferSize = strlen(_psSessionData->szTag) + 13;
   bufSend = malloc(nSendBufferSize + 1);
   memset(bufSend, 0, nSendBufferSize + 1);
@@ -313,80 +331,139 @@ int getAuthType(int hSocket, _MODULE_DATA* _psSessionData)
 
   if (medusaSend(hSocket, bufSend, strlen(bufSend), 0) < 0)
   {
-    writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
+    writeError(ERR_ERROR, "[%s] Failed: medusaSend was not successful", MODULE_NAME);
+    FREE(bufSend);
     return FAILURE;
   }
-  FREE(bufSend); 
- 
+  FREE(bufSend);
+
+  nReceiveBufferSize = 0;
   bufReceive = medusaReceiveRaw(hSocket, &nReceiveBufferSize);
   if (bufReceive == NULL)
   {
-    writeError(ERR_ERROR, "%s failed: medusaReceive returned no data. Exiting...", MODULE_NAME);
-    return FAILURE; 
+    writeError(ERR_ERROR, "[%s] Failed: medusaReceive returned no data.", MODULE_NAME);
+    return FAILURE;
+  }
+
+  /* If server supports STARTTLS, we use it */
+  if (strstr(bufReceive, "STARTTLS") != NULL)
+  {
+    FREE(bufReceive);
+
+    writeError(ERR_DEBUG_MODULE, "[%s] Initiating STARTTLS session.", MODULE_NAME);
+
+    bufSend = malloc(strlen(_psSessionData->szTag) + 11 + 1);
+    memset(bufSend, 0, strlen(_psSessionData->szTag) + 11 + 1);
+    sprintf(bufSend, "%s STARTTLS\r\n", _psSessionData->szTag);
+    if (medusaSend(hSocket, bufSend, strlen(bufSend), 0) < 0)
+    {
+      writeError(ERR_ERROR, "[%s] failed: medusaSend was not successful", MODULE_NAME);
+      FREE(bufSend);
+      return FAILURE;
+    }
+    FREE(bufSend);
+
+    nReceiveBufferSize = 0;
+    bufReceive = medusaReceiveRaw(hSocket, &nReceiveBufferSize);
+    if (bufReceive == NULL)
+    {
+      writeError(ERR_ERROR, "[%s] failed: medusaReceive returned no data.", MODULE_NAME);
+      return FAILURE;
+    }
+    /* OK Begin TLS negotiation now. */
+    else if (strstr(bufReceive, " OK ") != NULL)
+    {
+      FREE(bufReceive);
+
+      if (medusaConnectSocketSSL(params, hSocket) < 0)
+      {
+        writeError(ERR_ERROR, "[%s] Failed to establish SSLv3 connection.", MODULE_NAME);
+        return FAILURE;
+      }
+
+      /* Resend CAPABILITY request as the AUTH types may have changed. */
+      writeError(ERR_DEBUG_MODULE, "[%s] Sending IMAP CAPABILITIES request.", MODULE_NAME);
+      nSendBufferSize = strlen(_psSessionData->szTag) + 13;
+      bufSend = malloc(nSendBufferSize + 1);
+      memset(bufSend, 0, nSendBufferSize + 1);
+      sprintf(bufSend, "%s CAPABILITY\r\n", _psSessionData->szTag);
+
+      if (medusaSend(hSocket, bufSend, strlen(bufSend), 0) < 0)
+      {
+        writeError(ERR_ERROR, "[%s] Failed: medusaSend was not successful", MODULE_NAME);
+        FREE(bufSend);
+        return FAILURE;
+      }
+      FREE(bufSend);
+
+      nReceiveBufferSize = 0;
+      bufReceive = medusaReceiveRaw(hSocket, &nReceiveBufferSize);
+      if (bufReceive == NULL)
+      {
+        writeError(ERR_ERROR, "[%s] Failed: medusaReceive returned no data.", MODULE_NAME);
+        return FAILURE;
+      }
+    }
+  }
+
+  /* Process IMAP supported authentication types */
+  if (_psSessionData->nAuthType != AUTH_UNKNOWN)
+  {
+    writeError(ERR_DEBUG_MODULE, "[%s] Ignoring server requested AUTH type and using user-specified value.", MODULE_NAME);
   }
   else if ((strstr(bufReceive,"AUTH=LOGIN") != NULL))
   {
-    writeError(ERR_DEBUG_MODULE, "Detected authentication type: LOGIN");
+    writeError(ERR_DEBUG_MODULE, "Server requested authentication type: LOGIN");
     _psSessionData->nAuthType = AUTH_LOGIN;
   }
   else if ((strstr(bufReceive,"AUTH=PLAIN") != NULL))
   {
-    writeError(ERR_DEBUG_MODULE, "Detected authentication type: PLAIN");
+    writeError(ERR_DEBUG_MODULE, "Server requested authentication type: PLAIN");
     _psSessionData->nAuthType = AUTH_PLAIN;
   }
+  else if ((strstr(bufReceive,"AUTH=NTLM") != NULL))
+  {
+    writeError(ERR_DEBUG_MODULE, "Server requested authentication type: NTLM");
+    _psSessionData->nAuthType = AUTH_NTLM;
+  }
   else
   {
-    writeError(ERR_ERROR, "%s failed: Server did not respond that it supported either LOGIN or PLAIN as an authentication type. Use the AUTH module option to force the use of an authentication type.", MODULE_NAME);
+    writeError(ERR_ERROR, "[%s] Failed: Server did not respond that it supported any of the authentication types we handle (PLAIN, LOGIN, NTLM). Use the AUTH module option to force the use of an authentication type.", MODULE_NAME);
     return FAILURE; 
   }
 
+  FREE(bufReceive);
   return SUCCESS;
 }
 
-/* Stolen Right Out of Mutt */
-void mutt_to_base64 (unsigned char *out, unsigned char *in, size_t len, size_t olen)
+/* A0001 LOGIN username password */
+int sendAuthLogin(int hSocket, _MODULE_DATA* _psSessionData, char* szLogin, char* szPassword)
 {
-  int j;
+  unsigned char* bufSend = NULL;
+  unsigned char* szEncodedAuth = NULL;
+  int nSendBufferSize = 0;
+  int nRet = SUCCESS;
 
-  char B64Chars[64] = {
-    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
-    'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
-    'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
-    't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7',
-    '8', '9', '+', '/'
-  };
+  nSendBufferSize = strlen(_psSessionData->szTag) + 7 + strlen(szLogin) + 1 + strlen(szPassword) + 4 + 2; 
+  bufSend = malloc(nSendBufferSize + 1);
+  memset(bufSend, 0, nSendBufferSize + 1);
 
-  /* Back to Mutt code */
-  while (len >= 3 && olen > 10)
+  writeError(ERR_DEBUG_MODULE, "[%s] Sending authenticate login value: %s %s", MODULE_NAME, szLogin, szPassword); 
+  sprintf(bufSend, "%s LOGIN \"%s\" \"%s\"\r\n", _psSessionData->szTag, szLogin, szPassword);
+
+  if (medusaSend(hSocket, bufSend, strlen(bufSend), 0) < 0)
   {
-    *out++ = B64Chars[in[0] >> 2];
-    *out++ = B64Chars[((in[0] << 4) & 0x30) | (in[1] >> 4)];
-    *out++ = B64Chars[((in[1] << 2) & 0x3c) | (in[2] >> 6)];
-    *out++ = B64Chars[in[2] & 0x3f];
-    olen  -= 4;
-    len   -= 3;
-    in    += 3;
+    writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
+    nRet = FAILURE;
   }
 
-  /* clean up remainder */
-  if (len > 0 && olen > 4)
-  {
-    unsigned char fragment;
-
-    *out++ = B64Chars[in[0] >> 2];
-    fragment = (in[0] << 4) & 0x30;
-    if (len > 1)
-      fragment |= in[1] >> 4;
-    *out++ = B64Chars[fragment];
-    *out++ = (len < 2) ? '=' : B64Chars[(in[1] << 2) & 0x3c];
-    *out++ = '=';
-  }
-  *out = '\0';
+  FREE(bufSend); 
+  return(nRet);
 }
 
-int tryLogin(int hSocket, _MODULE_DATA* _psSessionData, sLogin** psLogin, char* szLogin, char* szPassword)
+/* A0001 AUTHENTICATE PLAIN credentials(base64) */
+int sendAuthPlain(int hSocket, _MODULE_DATA* _psSessionData, char* szLogin, char* szPassword)
 {
-  int iRet;
   unsigned char* bufSend = NULL;
   unsigned char* bufReceive = NULL;
   unsigned char* szTmp = NULL;
@@ -394,119 +471,287 @@ int tryLogin(int hSocket, _MODULE_DATA* _psSessionData, sLogin** psLogin, char*
   int nSendBufferSize = 0;
   int nReceiveBufferSize = 0;
 
-  /* A0001 LOGIN username password */
-  if (_psSessionData->nAuthType == AUTH_LOGIN)
-  {
-    nSendBufferSize = strlen(_psSessionData->szTag) + 7 + strlen(szLogin) + 1 + strlen(szPassword) + 4 + 2; 
-    bufSend = malloc(nSendBufferSize + 1);
-    memset(bufSend, 0, nSendBufferSize + 1);
-
-    writeError(ERR_DEBUG_MODULE, "[%s] Sending authenticate login value: %s %s", MODULE_NAME, szLogin, szPassword); 
-    sprintf(bufSend, "%s LOGIN \"%s\" \"%s\"\r\n", _psSessionData->szTag, szLogin, szPassword);
-  }
-  /* A0001 AUTHENTICATE PLAIN credentials(base64) */
-  else if (_psSessionData->nAuthType == AUTH_PLAIN)
-  {
-    /* Send initial AUTHENTICATE PLAIN command */
-    nSendBufferSize = strlen(_psSessionData->szTag) + 21;
-    bufSend = malloc(nSendBufferSize + 1);
-    memset(bufSend, 0, nSendBufferSize + 1);
-    sprintf(bufSend, "%s AUTHENTICATE PLAIN\r\n", _psSessionData->szTag);
-    
-    if (medusaSend(hSocket, bufSend, strlen(bufSend), 0) < 0)
-    {
-      writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
-    }
-    FREE(bufSend); 
- 
-    nReceiveBufferSize = 0;
-    bufReceive = medusaReceiveRaw(hSocket, &nReceiveBufferSize);
-    if (bufReceive == NULL)
-    {
-      writeError(ERR_ERROR, "%s failed: medusaReceive returned no data.", MODULE_NAME);
-      (*psLogin)->iResult = LOGIN_RESULT_ERROR;
-      iRet = MSTATE_EXITING;
-      return(iRet);
-    }
-    else if (strstr(bufReceive,"+") == NULL) 
-    {
-      writeError(ERR_ERROR, "[%s] IMAP server did not respond with \"+\" to AUTHENTICATE PLAIN request.", MODULE_NAME);
-      (*psLogin)->iResult = LOGIN_RESULT_ERROR;
-      iRet = MSTATE_EXITING;
-      return(iRet);
-    }
-
-    /* Build a null separated string of szLogin \0 szLogin \0 szPassword */
-    nSendBufferSize = strlen(szLogin) + 1 + strlen(szLogin) + 1 + strlen(szPassword); 
-    
-    szTmp = malloc(nSendBufferSize + 1);
-    memset(szTmp, 0, nSendBufferSize + 1);
-    
-    /* username\0username\0password */
-    memcpy(szTmp, szLogin, strlen(szLogin));
-    memcpy(szTmp + strlen(szLogin) + 1, szLogin, strlen(szLogin));
-    memcpy(szTmp + strlen(szLogin) + 1 + strlen(szLogin) + 1, szPassword, strlen(szPassword)); 
-    
-    szEncodedAuth = malloc(2 * nSendBufferSize + 1);
-    memset(szEncodedAuth, 0, 2 * nSendBufferSize + 1);
-   
-    mutt_to_base64(szEncodedAuth, szTmp, nSendBufferSize, 2 * nSendBufferSize);
-    FREE(szTmp);
-   
-    writeError(ERR_DEBUG_MODULE, "[%s] Sending authenticate plain value: %s", MODULE_NAME, szEncodedAuth); 
-    nSendBufferSize = strlen(szEncodedAuth) + 2;
-    bufSend = malloc(nSendBufferSize + 1);
-    memset(bufSend, 0, nSendBufferSize + 1);
-    sprintf(bufSend, "%s\r\n", szEncodedAuth);
-    FREE(szEncodedAuth);
-  }
-  else
-  {
-    writeError(ERR_ERROR, "[%s] Unknown authentication type.", MODULE_NAME);
-    return FAILURE;
-  }
-
+  /* Send initial AUTHENTICATE PLAIN command */
+  nSendBufferSize = strlen(_psSessionData->szTag) + 21;
+  bufSend = malloc(nSendBufferSize + 1);
+  memset(bufSend, 0, nSendBufferSize + 1);
+  sprintf(bufSend, "%s AUTHENTICATE PLAIN\r\n", _psSessionData->szTag);
+  
   if (medusaSend(hSocket, bufSend, strlen(bufSend), 0) < 0)
   {
     writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
   }
   FREE(bufSend); 
+
+  nReceiveBufferSize = 0;
+  bufReceive = medusaReceiveRaw(hSocket, &nReceiveBufferSize);
+  if (bufReceive == NULL)
+  {
+    writeError(ERR_ERROR, "%s failed: medusaReceive returned no data.", MODULE_NAME);
+    return FAILURE;
+  }
+  else if (strstr(bufReceive,"+") == NULL) 
+  {
+    writeError(ERR_ERROR, "[%s] IMAP server did not respond with \"+\" to AUTHENTICATE PLAIN request.", MODULE_NAME);
+    writeError(ERR_ERROR, "[%s] IMAP server sent the following response: %s", MODULE_NAME, bufReceive);
+    return FAILURE;
+  }
+
+  /* Build a null separated string of szLogin \0 szLogin \0 szPassword */
+  nSendBufferSize = strlen(szLogin) + 1 + strlen(szLogin) + 1 + strlen(szPassword); 
+  
+  szTmp = malloc(nSendBufferSize + 1);
+  memset(szTmp, 0, nSendBufferSize + 1);
+  
+  /* username\0username\0password */
+  memcpy(szTmp, szLogin, strlen(szLogin));
+  memcpy(szTmp + strlen(szLogin) + 1, szLogin, strlen(szLogin));
+  memcpy(szTmp + strlen(szLogin) + 1 + strlen(szLogin) + 1, szPassword, strlen(szPassword)); 
+  
+  szEncodedAuth = malloc(2 * nSendBufferSize + 1);
+  memset(szEncodedAuth, 0, 2 * nSendBufferSize + 1);
+  base64_encode(szTmp, nSendBufferSize, szEncodedAuth);
+  FREE(szTmp);
  
+  writeError(ERR_DEBUG_MODULE, "[%s] Sending authenticate plain value: %s", MODULE_NAME, szEncodedAuth); 
+  nSendBufferSize = strlen(szEncodedAuth) + 2;
+  bufSend = malloc(nSendBufferSize + 1);
+  memset(bufSend, 0, nSendBufferSize + 1);
+  sprintf(bufSend, "%s\r\n", szEncodedAuth);
+  FREE(szEncodedAuth);
+  
+  if (medusaSend(hSocket, bufSend, strlen(bufSend), 0) < 0)
+  {
+    writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
+    return FAILURE;
+  }
+
+  FREE(bufSend); 
+  return SUCCESS;
+}
+
+/*
+  A0001 AUTHENTICATE NTLM
+  
+  NTLM IMAP Authentication
+  Based on:
+    http://curl.haxx.se/rfc/ntlm.html#ntlmImapAuthentication
+    http://src.opensolaris.org/source/xref/sfw/usr/src/cmd/fetchmail/fetchmail-6.3.8/README.NTLM
+*/
+int sendAuthNTLM(int hSocket, _MODULE_DATA* _psSessionData, char* szLogin, char* szPassword)
+{
+  unsigned char* bufSend = NULL;
+  unsigned char* bufReceive = NULL;
+  unsigned char* szTmpBuf = NULL;
+  unsigned char* szTmpBuf64 = NULL;
+  unsigned char* szTmpChall = NULL;
+  int nSendBufferSize = 0;
+  int nReceiveBufferSize = 0;
+
+  /* --- Send initial AUTHENTICATE NTLM command --- */
+  nSendBufferSize = strlen(_psSessionData->szTag) + 21;
+  bufSend = malloc(nSendBufferSize + 1);
+  memset(bufSend, 0, nSendBufferSize + 1);
+  sprintf(bufSend, "%s AUTHENTICATE NTLM\r\n", _psSessionData->szTag);
+  
+  if (medusaSend(hSocket, bufSend, strlen(bufSend), 0) < 0)
+  {
+    writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
+  }
+  FREE(bufSend); 
+
+  /* Server should respond with an empty challenge, consisting simply of a "+" */
+  nReceiveBufferSize = 0;
+  bufReceive = medusaReceiveRaw(hSocket, &nReceiveBufferSize);
+  if (bufReceive == NULL)
+  {
+    writeError(ERR_ERROR, "%s failed: medusaReceive returned no data.", MODULE_NAME);
+    return FAILURE;
+  }
+  else if (strstr(bufReceive, "+") == NULL) 
+  {
+    writeError(ERR_ERROR, "[%s] IMAP server did not respond with \"+\" to AUTHENTICATE NTLM request.", MODULE_NAME);
+    return FAILURE;
+  }
+  FREE(bufReceive);
+
+  /* --- Send Base-64 encoded Type-1 message --- */
+  szTmpBuf = malloc( sizeof(tSmbNtlmAuthRequest) );  
+  memset(szTmpBuf, 0, sizeof(tSmbNtlmAuthRequest));
+
+  szTmpBuf64 = malloc( 2 * sizeof(tSmbNtlmAuthRequest) + 2);  
+  memset(szTmpBuf64, 0, 2 * sizeof(tSmbNtlmAuthRequest) + 2);
+
+  buildAuthRequest((tSmbNtlmAuthRequest*)szTmpBuf, 0, NULL, NULL);
+  base64_encode(szTmpBuf, sizeof(tSmbNtlmAuthResponse), szTmpBuf64);
+
+  nSendBufferSize = strlen(szTmpBuf64) + 2;
+  bufSend = malloc(nSendBufferSize + 1);
+  memset(bufSend, 0, nSendBufferSize + 1);
+  sprintf(bufSend, "%s\r\n", szTmpBuf64);
+
+  FREE(szTmpBuf);
+  FREE(szTmpBuf64);
+
+  if (medusaSend(hSocket, bufSend, strlen(bufSend), 0) < 0)
+  {
+    writeError(ERR_ERROR, "[%s] failed: medusaSend was not successful", MODULE_NAME);
+    return FAILURE;
+  }
+  FREE(bufSend); 
+  
+  /* Server should respond with a Base-64 encoded Type-2 challenge message. The challenge response format is 
+     specified by RFC 1730 ("+", followed by a space, followed by the challenge message). */
+  nReceiveBufferSize = 0;
+  bufReceive = medusaReceiveRaw(hSocket, &nReceiveBufferSize);
+  if (bufReceive == NULL)
+  {
+    writeError(ERR_ERROR, "[%s] failed: medusaReceive returned no data.", MODULE_NAME);
+    return FAILURE;
+  }
+  else if (strncmp(bufReceive, "+ ", 2) != 0)
+  {
+    writeError(ERR_ERROR, "[%s] Server did not send valid Type-2 challenge response.", MODULE_NAME);
+    return FAILURE;
+  }
+
+  szTmpBuf = ((char*)index(bufReceive, '\r'));
+  szTmpBuf[0] = '\0';
+
+  szTmpChall = malloc( sizeof(tSmbNtlmAuthChallenge) );  
+  memset(szTmpChall, 0, sizeof(tSmbNtlmAuthChallenge));
+  
+  base64_decode(bufReceive + 2, szTmpChall);
+
+  FREE(bufReceive);
+  
+  /* --- Calculate and send Base-64 encoded Type 3 response --- */
+  szTmpBuf = malloc( sizeof(tSmbNtlmAuthResponse) );
+  memset(szTmpBuf, 0, sizeof(tSmbNtlmAuthResponse));
+  
+  szTmpBuf64 = malloc( 2 * sizeof(tSmbNtlmAuthResponse) + 2);
+  memset(szTmpBuf64, 0, 2 * sizeof(tSmbNtlmAuthResponse) + 2);
+
+  buildAuthResponse((tSmbNtlmAuthChallenge*)szTmpChall, (tSmbNtlmAuthResponse*)szTmpBuf, 0, szLogin, szPassword, NULL, NULL);
+  base64_encode(szTmpBuf, sizeof(tSmbNtlmAuthResponse), szTmpBuf64);
+
+  nSendBufferSize = strlen(szTmpBuf64) + 2;
+  bufSend = malloc(nSendBufferSize + 1);
+  memset(bufSend, 0, nSendBufferSize + 1);
+  sprintf(bufSend, "%s\r\n", szTmpBuf64);
+
+  if (medusaSend(hSocket, bufSend, nSendBufferSize, 0) < 0)
+  {
+    writeError(ERR_ERROR, "[%s] failed: medusaSend was not successful", MODULE_NAME);
+    return FAILURE;
+  }
+
+  FREE(szTmpBuf);
+  FREE(szTmpBuf64);
+  FREE(szTmpChall);
+  FREE(bufSend);
+
+  /* Server should validate the response and indicate the result of authentication.
+     e.g.  0001 OK AUTHENTICATE NTLM completed. */
+
+  return SUCCESS;
+}
+
+int tryLogin(int hSocket, _MODULE_DATA* _psSessionData, sLogin** psLogin, char* szLogin, char* szPassword)
+{
+  int nRet;
+  unsigned char* bufReceive = NULL;
+  int nReceiveBufferSize = 0;
+
+  switch(_psSessionData->nAuthType)
+  {
+    case AUTH_LOGIN:
+      writeError(ERR_DEBUG_MODULE, "[%s] Sending LOGIN Authentication.", MODULE_NAME);
+      nRet = sendAuthLogin(hSocket, _psSessionData, szLogin, szPassword);
+      break;
+    case AUTH_PLAIN:
+      writeError(ERR_DEBUG_MODULE, "[%s] Sending PLAIN Authentication.", MODULE_NAME);
+      nRet = sendAuthPlain(hSocket, _psSessionData, szLogin, szPassword);
+      break;
+    case AUTH_NTLM:
+      writeError(ERR_DEBUG_MODULE, "[%s] Sending NTLM Authentication.", MODULE_NAME);
+      nRet = sendAuthNTLM(hSocket, _psSessionData, szLogin, szPassword);
+      break;
+    default:
+      break;
+  }
+
+  if (nRet == FAILURE)
+  {
+    writeError(ERR_ERROR, "[%s] Failed during sending of authentication data.", MODULE_NAME);
+    (*psLogin)->iResult = LOGIN_RESULT_UNKNOWN;
+    setPassResult(*psLogin, szPassword);
+    return MSTATE_EXITING;
+  }
+
+  /*
+    Exchange 2003 Server Response Messages
+
+    NO Logon failure: unknown user name or bad password.
+    NO The specified authentication package is not supported.
+    NO Clear text passwords have been disabled for this protocol.
+    NO Cleartext login on this server requires the use of transport level security (SSL/TLS)
+  */
+ 
+  writeError(ERR_DEBUG_MODULE, "[%s] Retrieving server response.", MODULE_NAME);
   nReceiveBufferSize = 0;
   bufReceive = medusaReceiveRaw(hSocket, &nReceiveBufferSize);
   if (bufReceive == NULL)
   {
     writeError(ERR_ERROR, "%s failed: medusaReceive returned no data.", MODULE_NAME);
     (*psLogin)->iResult = LOGIN_RESULT_ERROR;
-    iRet = MSTATE_EXITING;
+    nRet = MSTATE_EXITING;
   }
   else if (strstr(bufReceive,"OK") != NULL) 
   {
     writeError(ERR_DEBUG_MODULE, "[%s] Login attempt successful.", MODULE_NAME);
     (*psLogin)->iResult = LOGIN_RESULT_SUCCESS;
-    iRet = MSTATE_EXITING;
+    nRet = MSTATE_EXITING;
+  }
+  else if (strstr(bufReceive,"NO Clear text passwords have been disabled for this protocol.") != NULL) 
+  {
+    writeError(ERR_ERROR, "[%s] Server reports that clear-text passwords have been disabled.", MODULE_NAME);
+    (*psLogin)->iResult = LOGIN_RESULT_ERROR;
+    nRet = MSTATE_EXITING;
+  }
+  else if (strstr(bufReceive,"NO Cleartext login on this server requires the use of transport level security (SSL/TLS)") != NULL) 
+  {
+    writeError(ERR_ERROR, "[%s] Server reports that clear-text passwords are only allowed over SSL/TLS.", MODULE_NAME);
+    (*psLogin)->iResult = LOGIN_RESULT_ERROR;
+    nRet = MSTATE_EXITING;
+  }
+  else if (strstr(bufReceive,"NO The specified authentication package is not supported.") != NULL) 
+  {
+    writeError(ERR_ERROR, "[%s] Server reports that the specified authentication package is not supported.", MODULE_NAME);
+    (*psLogin)->iResult = LOGIN_RESULT_ERROR;
+    nRet = MSTATE_EXITING;
   }
   else if (strstr(bufReceive,"NO") != NULL) 
   {
     writeError(ERR_DEBUG_MODULE, "[%s] Login attempt failed.", MODULE_NAME);
     (*psLogin)->iResult = LOGIN_RESULT_FAIL;
-    iRet = MSTATE_RUNNING;
+    nRet = MSTATE_RUNNING;
   }
   else if (strstr(bufReceive,"BAD") != NULL) 
   {
     writeError(ERR_ERROR, "[%s] IMAP server responded that the command was unknown or the arguments were invalid.", MODULE_NAME);
     (*psLogin)->iResult = LOGIN_RESULT_ERROR;
-    iRet = MSTATE_EXITING;
+    nRet = MSTATE_EXITING;
   }
   else 
   {
-    writeError(ERR_ERROR, "[%s] Unknown IMAP server response.", MODULE_NAME);
+    writeError(ERR_ERROR, "[%s] Unknown IMAP server response: %s", MODULE_NAME, bufReceive);
     (*psLogin)->iResult = LOGIN_RESULT_ERROR;
-    iRet = MSTATE_EXITING;
+    nRet = MSTATE_EXITING;
   }
 
   FREE(bufReceive);
   setPassResult((*psLogin), szPassword);
 
-  return(iRet);
+  return(nRet);
 }
