@@ -25,6 +25,8 @@
 **
 **    MySQL <4.1 Pass Hash Support: jmk <jmk@foofus.net>
 **
+**    This module supports MySQL protocol 10.
+**
 **    The older MySQL pre-4.1 authentication scheme is vulnerable to a pass-the-
 **    hash authentication attack. Utilizing the old-style hashes gathered from a
 **    MySQL database, a user can use Medusa to verify their validity on other 
@@ -45,7 +47,7 @@
 #define MODULE_AUTHOR  "JoMo-Kun <jmk@foofus.net>"
 #define MODULE_SUMMARY_USAGE  "Brute force module for MySQL sessions"
 #define MODULE_VERSION    "1.2"
-#define MODULE_VERSION_SVN "$Id: mysql.c 811 2007-08-28 21:19:29Z jmk $"
+#define MODULE_VERSION_SVN "$Id: mysql.c 1052 2009-01-21 22:27:47Z jmk $"
 #define MODULE_SUMMARY_FORMAT  "%s : version %s"
 #define BUF_SIZE 300
 
@@ -56,8 +58,9 @@
         }
 
 #define PORT_MYSQL 3306
-#define PROTO_NEW 0
+#define PROTO_UNKNOWN 0
 #define PROTO_OLD 1
+#define PROTO_NEW 2
 #define PASSWORD 1
 #define HASH 2
 
@@ -396,8 +399,7 @@ void scramble_323(char *to, _MYSQL_DATA *_psSessionData, const char *message, co
       hash_password(hash_pass, password, strlen(password));
 
     hash_password(hash_message, message, SCRAMBLE_LENGTH_323);
-    randominit(&rand_st,hash_pass[0] ^ hash_message[0],
-               hash_pass[1] ^ hash_message[1]);
+    randominit(&rand_st, hash_pass[0] ^ hash_message[0], hash_pass[1] ^ hash_message[1]);
     for (; message < message_end; message++)
       *to++= (char) (floor(my_rnd(&rand_st)*31)+64);
     extra=(char) (floor(my_rnd(&rand_st)*31));
@@ -444,124 +446,126 @@ void scramble(char *to, _MYSQL_DATA *_psSessionData, const char *message, const
 
 int MySQLPrepareAuthOld(_MYSQL_DATA *_psSessionData, char* szLogin, char* szPassword, char* szSessionSalt, unsigned char** szResponse, unsigned long* iResponseLength)
 {
-  char* szHash;
-  unsigned long login_len = strlen(szLogin) > 32 ? 32 : strlen(szLogin);
-  unsigned long response_len = 4 /* header */  +
-                               2 /* client flags */  +
-                               3 /* max packet len */  +
-                               login_len + 1 + 8 /* scrambled password len */ ;
   unsigned char* response;
+  /* usernames limited to 16 characters - http://dev.mysql.com/doc/refman/5.1/en/user-names.html */
+  int login_len = strlen(szLogin) > 16 ? 16 : strlen(szLogin);
+  int response_len = 4 /* header */  +
+                     2 /* client flags */  +
+                     3 /* max packet len */  +
+                     login_len /* username length */ + 
+                     1 /* NULL terminate username */ + 
+                     8; /* scrambled password len */
 
-  *(iResponseLength) = response_len;
 
-  /* < + 15 causes FREE issue. was 4 in original code. ??? */
-  response = (unsigned char *) malloc(response_len + 15);
-  if (response == NULL) {
-    writeError(ERR_ERROR, "%s: Error allocating memory.", MODULE_NAME);
-    return FAILURE;  
-  }
+  response = (unsigned char *) malloc(response_len + 1);
+  memset(response, 0, response_len + 1);
 
-  memset(response, 0, response_len + 15);
+  response[0] = response_len - 4;   /* packet length */
+  response[1] = 0x00;
+  response[2] = 0x00;
+  response[3] = 0x01;               /* packet number */
+
+  response[4] = 0x85;               /* client flag */
+  response[5] = 0x24;
+
+  response[6] = 0x00;               /* max packet */
+  response[7] = 0x00;
+  response[8] = 0x00;
+  
+  strncpy(response + 9, szLogin, login_len);
 
-  *((unsigned long *) response) = response_len - 4;  /* packet length */
-  response[3] = 0x01;           /* packet number */
-  response[4] = 0x85;           /* client flag */
-  response[5] = 0x24;           /* client flag */
-  response[6] = 0x00;           /* max packet */
-  response[7] = 0x00;           /* max packet */
-  response[8] = 0x00;           /* max packet */
-  strncpy(response + 9, szLogin, 32);
   response[9 + login_len] = '\0';       /* null terminate login */
   
-  if ( strcmp(szPassword, "") )
+  if ( strcmp(szPassword, "") != 0) /* password set */
     scramble_323((char *) &response[9 + login_len + 1], _psSessionData, szSessionSalt, szPassword);
 
+  *(iResponseLength) = response_len;
   *szResponse = response;
   return SUCCESS;
 }
 
-/* This is a weird packet format used on a rare case when the newer Mysql
-   still has a password table entry in the old shorter format */
+/* Protocol 10 is used by MySQL 3.22 and later. However, MySQL 4.1 introduced a new password algorithm.
+   In some cases, MySQL 4.1 and later systems will contain accounts which are still configured with password
+   hashes generated using the older algorithm. When we authenticate to a 4.1 server and this is the case,
+   the server is nice enough to tell us and allow us to reauthenticate. This function generates the appropriate
+   response for this particular case.
+*/
 int MySQLPrepareAuthNewOld(_MYSQL_DATA *_psSessionData, char* szLogin, char* szPassword, char* szSessionSalt, unsigned char** szResponse, unsigned long* iResponseLength)
 {
-  char* szHash;
-  unsigned long login_len = strlen(szLogin) > 32 ? 32 : strlen(szLogin);
-  unsigned long response_len = 4 /* header */  +
-                               1 + 8 /* scrambled password len */ ;
   unsigned char* response;
+  /* usernames limited to 16 characters - http://dev.mysql.com/doc/refman/5.1/en/user-names.html */
+  int login_len = strlen(szLogin) > 16 ? 16 : strlen(szLogin);
+  int response_len = 4 +  /* header */
+                     1 +
+                     8; /* scrambled password length */
 
-  *(iResponseLength) = response_len;
+  response = (unsigned char *) malloc(response_len + 1);
+  memset(response, 0, response_len + 1);
 
-  /* < + 15 causes FREE issue. was 4 in original code. ??? */
-  response = (unsigned char *) malloc(response_len + 15);
-  if (response == NULL) {
-    writeError(ERR_ERROR, "%s: Error allocating memory.", MODULE_NAME);
-    return FAILURE;  
-  }
-
-  memset(response, 0, response_len + 15);
-
-  *((unsigned long *) response) = response_len - 4;  /* packet length */
-  response[3] = 0x03;           /* packet number */
+  response[0] = response_len - 4;   /* packet length */
+  response[3] = 0x03;               /* packet number */
   scramble_323((char *) &response[4], _psSessionData, szSessionSalt, szPassword);
 
+  *(iResponseLength) = response_len;
   *szResponse = response;
   return SUCCESS;
 }
 
+/* http://www.redferni.uklinux.net/mysql/MySQL-Protocol.html */ 
 int MySQLPrepareAuth(_MYSQL_DATA *_psSessionData, char* szLogin, char* szPassword, char* szSessionSalt, unsigned char** szResponse, unsigned long* iResponseLength)
 {
-  char* szHash;
-  unsigned long login_len = strlen(szLogin) > 32 ? 32 : strlen(szLogin);
-  unsigned long response_len = 4 /* header */  +
-                               2 /* client flags */  +
-                               3 /* max packet len */  +
-                               login_len + 1 + 48 /* scrambled password len */ ;
   unsigned char* response;
+  /* usernames limited to 16 characters - http://dev.mysql.com/doc/refman/5.1/en/user-names.html */
+  int login_len = strlen(szLogin) > 16 ? 16 : strlen(szLogin);
+  int response_len = 4  /* header */  +
+                     4  /* client flags */  +
+                     4  /* max packet len */  +
+                     1  /* charset */ +
+                     23 /* future expansion */ +
+                     login_len /* username */ +
+                     1 /* NULL termination */ + 
+                     1; /* password length */
+    
+  if ( strcmp(szPassword, "") != 0 ) /* password set */
+    response_len += 20;
 
-  *(iResponseLength) = response_len; 
+  response = (unsigned char *) malloc(response_len + 1);
+  memset(response, 0, response_len + 1);
 
-  /* < + 15 causes FREE issue. was 4 in original code. ??? */
-  response = (unsigned char *) malloc(response_len + 15);
-  if (response == NULL) {
-    writeError(ERR_ERROR, "%s: Error allocating memory.", MODULE_NAME);
-    return FAILURE;
-  }
-
-  /* Much of this derived from 
-    http://www.redferni.uklinux.net/mysql/MySQL-Protocol.html 
-  */
-
-  memset(response, 0, response_len + 15);
-
-  *((unsigned long *) response) = response_len - 4;  /* packet length */
-  response[3] = 0x01;           /* packet number */
-  /* response[4] = 0x85;        */    /* client flag */
-  response[4] = 0x05;           /* client flag */
-  /* response[5] = 0x24; */           /* client flag */ 
-  response[5] = 0xa6;           /* client flag */ 
-  response[6] = 0x03;           /* max packet */
-  response[7] = 0x00;           /* max packet */        
-  response[8] = 0x00;           /* max packet */
-
-  response[9] = 0x00;    /* Username is null in a 4.1 auth packet */
+  response[0] = response_len - 4;     /* packet body length - exclude header */
+  response[1] = 0x00;
+  response[2] = 0x00;
+  response[3] = 0x01;                 /* packet number */
+  
+  //response[4] = 0x85;               /* client flag */
+  response[4] = 0x05;                 /* client flag */
+  //response[5] = 0x24;
+  response[5] = 0xa6;
+  response[6] = 0x03;
+  response[7] = 0x00;
+  
+  response[8] = 0x00;                 /* max packet */
+  response[9] = 0x00;
   response[10] = 0x00;    
   response[11] = 0x01;    
-  response[12] = 0x08;    /* Charset Latin */
+  
+  response[12] = 0x21;                /* charset utf8 */
 
-  /* OK, now add the username null terminated */
+  strncpy(response + 36, szLogin, login_len);  /* NULL terminated username */
 
-  strncpy(response + 36, szLogin, 32);
-
-  if ( strcmp(szPassword, "") )
+  if ( strcmp(szPassword, "") == 0 ) /* no password set */
   {
-/* Now add a 0x14 which is always the scrambled length of the password (20) */ 
-    response[36+strlen(szLogin)+1] = 0x14;
+    response[36 + login_len + 1] = 0x00;
+  }
+  else
+  {
+    response[36 + login_len + 1] = 0x14;  /* set length of scrambled password - 0x14 (20) */
 
-    /* Now add the scramble SHA hash */
-    scramble((char *) &response[36+strlen(szLogin)+2], _psSessionData, szSessionSalt, szPassword); 
+    /* generate SHA password hash */
+    scramble((char *) &response[36 + login_len + 1 + 1], _psSessionData, szSessionSalt, szPassword); 
   }
   
+  *(iResponseLength) = response_len; 
   *szResponse = response;
   return SUCCESS;
 }
@@ -599,14 +603,13 @@ int MySQLSessionInit(int hSocket, unsigned char** szSessionSalt)
 
   if (bufReceive[4] < 10)
   {
-    writeError(ERR_ERROR, "%s: Server responded requestion protocol version (%d). Version 10 support required." MODULE_NAME, bufReceive[4]);
+    writeError(ERR_ERROR, "%s: Server responded requestion protocol version (%d). Version 10 support required.", MODULE_NAME, bufReceive[4]);
     FREE(bufReceive);
     return FAILURE;
   }
-
-  if (bufReceive[4] > 10)
+  else if (bufReceive[4] > 10)
   {
-    writeError(ERR_WARNING, "%s: Server responded requestion protocol version (%d). Support for versions >10 is unknown." MODULE_NAME, bufReceive[4]);
+    writeError(ERR_WARNING, "%s: Server responded requestion protocol version (%d). Support for versions >10 is unknown.", MODULE_NAME, bufReceive[4]);
   }
 
   /* check server version */
@@ -614,7 +617,7 @@ int MySQLSessionInit(int hSocket, unsigned char** szSessionSalt)
 
   if (!(strstr(szServerVersion, "3.") || strstr(szServerVersion, "4.") || strstr(szServerVersion, "5.") ))
   {
-    writeError(ERR_ERROR, "%s: Server responded requestion version (%d). Only versions 3.x 4.x 5.x currently supported." MODULE_NAME, szServerVersion);
+    writeError(ERR_ERROR, "%s: Server responded requestion version (%d). Only versions 3.x 4.x 5.x currently supported.", MODULE_NAME, szServerVersion);
     FREE(bufReceive);
     return FAILURE;
   }
@@ -635,7 +638,7 @@ int MySQLSessionInit(int hSocket, unsigned char** szSessionSalt)
 
     if (strlen(*szSessionSalt) != 20)
     {
-      writeError(ERR_ERROR, "%s: Failed to retrieve valid session salt." MODULE_NAME);
+      writeError(ERR_ERROR, "%s: Failed to retrieve valid session salt.", MODULE_NAME);
       FREE(bufReceive);
       return FAILURE;
     }
@@ -652,7 +655,7 @@ int MySQLSessionInit(int hSocket, unsigned char** szSessionSalt)
     memcpy(*szSessionSalt, bufReceive + strlen(szServerVersion) + 10, 9);
 
     if (strlen(*szSessionSalt) != 8) {
-      writeError(ERR_ERROR, "%s: Failed to retrieve valid session salt." MODULE_NAME);
+      writeError(ERR_ERROR, "%s: Failed to retrieve valid session salt.", MODULE_NAME);
       FREE(bufReceive);
       return FAILURE;
     }
@@ -669,7 +672,6 @@ int MySQLSessionInit(int hSocket, unsigned char** szSessionSalt)
 int tryLogin(int hSocket, sLogin** psLogin, _MYSQL_DATA* _psSessionData, char* szLogin, char* szPassword)
 {
   int iRet, iReturnCode;
-  unsigned char bufSend[BUF_SIZE];
   unsigned char* bufReceive = NULL;
   unsigned char* szSessionSalt = NULL;
   unsigned char* szResponse = NULL;
@@ -680,7 +682,7 @@ int tryLogin(int hSocket, sLogin** psLogin, _MYSQL_DATA* _psSessionData, char* s
   iRet = MySQLSessionInit(hSocket, &szSessionSalt);
   if (iRet == FAILURE)
   {
-    writeError(ERR_ERROR, "%s: Failed to initialize MySQL connection (%s).", MODULE_NAME, (*psLogin)->psServer->pHostIP);
+    writeError(ERR_ERROR, "[%s] Failed to initialize MySQL connection (%s).", MODULE_NAME, (*psLogin)->psServer->pHostIP);
     (*psLogin)->iResult = LOGIN_RESULT_ERROR;
     (*psLogin)->iStatus = LOGIN_FAILED;
     return MSTATE_EXITING;
@@ -690,13 +692,13 @@ int tryLogin(int hSocket, sLogin** psLogin, _MYSQL_DATA* _psSessionData, char* s
   if (strlen(szSessionSalt) == 8 || _psSessionData->protoFlag == PROTO_OLD)
   {
     if (_psSessionData->protoFlag == PROTO_OLD) {
-      writeError(ERR_DEBUG, "%s: using older auth per previous attempt response.", MODULE_NAME);
+      writeError(ERR_DEBUG, "[%s] Using older style authentication based on previous server response.", MODULE_NAME);
     }
     
     iRet = MySQLPrepareAuthOld(_psSessionData, szLogin, szPassword, szSessionSalt, &szResponse, &iResponseLength);
     if (iRet == FAILURE)
     {
-      writeError(ERR_ERROR, "%s: Failed to create client authentication packet.", MODULE_NAME);
+      writeError(ERR_ERROR, "[%s] Failed to create client authentication packet.", MODULE_NAME);
       return FAILURE;
     }
   } 
@@ -717,7 +719,6 @@ int tryLogin(int hSocket, sLogin** psLogin, _MYSQL_DATA* _psSessionData, char* s
     FREE(szResponse);
     return FAILURE;
   }
-
   FREE(szResponse);
 
   /* process authentication response */
@@ -740,7 +741,7 @@ int tryLogin(int hSocket, sLogin** psLogin, _MYSQL_DATA* _psSessionData, char* s
     
     if (bufReceive[5] == 0xe3 && bufReceive[6] == 0x04)
     {
-      writeError(ERR_ERROR, "%s failed: MYSQL VERSION IS NEWER\n", MODULE_NAME);
+      writeError(ERR_ERROR, "[%s] failed: MYSQL VERSION IS NEWER\n", MODULE_NAME);
       (*psLogin)->iResult = LOGIN_RESULT_ERROR;
       (*psLogin)->iStatus = LOGIN_FAILED;
       iReturnCode = MSTATE_EXITING;
@@ -750,33 +751,36 @@ int tryLogin(int hSocket, sLogin** psLogin, _MYSQL_DATA* _psSessionData, char* s
   }
   else if (bufReceive[4] == 0xFE)
   {
-    writeError(ERR_DEBUG, "%s failed: Server responded resend older Auth request.\n", MODULE_NAME);
+    /* Protocol 10 is used by MySQL 3.22 and later. However, MySQL 4.1 introduced a new password algorithm.
+       In some cases, MySQL 4.1 and later systems will contain accounts which are still configured with password
+       hashes generated using the older algorithm. When we authenticate to a 4.1 server and this is the case,
+       the server is nice enough to tell us and allow us to reauthenticate.
+    */
+
+    writeError(ERR_DEBUG, "[%s] Server requested older authentication type. It is likely the remote account exists and has an older style password hash.", MODULE_NAME);
     (*psLogin)->iResult = LOGIN_RESULT_FAIL;
 
-    /* OK, so what the blank do I do now? */
-
+    /* Attempt authentication again using old-style password hash and existing connection */
     _psSessionData->protoFlag = PROTO_OLD;
-
-    /* Well, you gotta send just the old style scramble using the existing connection */
-
-    iRet = MySQLPrepareAuthNewOld(_psSessionData, szLogin, szPassword, szSessionSalt, &szResponse,&iResponseLength);
+    
+    iRet = MySQLPrepareAuthNewOld(_psSessionData, szLogin, szPassword, szSessionSalt, &szResponse, &iResponseLength);
     if (iRet == FAILURE)
     {
-      writeError(ERR_ERROR, "%s: Failed to create client authentication packet.", MODULE_NAME);
+      writeError(ERR_ERROR, "[%s] Failed to create client authentication packet.", MODULE_NAME);
       return FAILURE;
     }
 
     /* send authentication attempt */
     if (medusaSend(hSocket, szResponse, iResponseLength, 0) < 0)
     {
-      writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
+      writeError(ERR_ERROR, "[%s] medusaSend was not successful", MODULE_NAME);
       FREE(szResponse);
       return FAILURE;
     }
-
     FREE(szResponse);
 
     /* process authentication response */
+    FREE(bufReceive);
     nReceiveBufferSize = 0;
     bufReceive = medusaReceiveRaw(hSocket, &nReceiveBufferSize);
     if (bufReceive == NULL)
