@@ -36,7 +36,7 @@
 #define MODULE_AUTHOR  "JoMo-Kun <jmk@foofus.net>"
 #define MODULE_SUMMARY_USAGE  "Brute force module for SSH v2 sessions"
 #define MODULE_VERSION    "1.0.2"
-#define MODULE_VERSION_SVN "$Id: ssh.c 791 2007-06-26 20:57:18Z jmk $"
+#define MODULE_VERSION_SVN "$Id: ssh.c 955 2008-08-20 16:04:44Z jmk $"
 #define MODULE_SUMMARY_FORMAT  "%s : version %s"
 #define MODULE_SUMMARY_FORMAT_WARN  "%s : version %s (%s)"
 #define LIBSSH2_WARNING "No usable LIBSSH2. Module disabled."
@@ -50,14 +50,19 @@
 #define SSH_AUTH_KBDINT 2
 #define SSH_AUTH_PASSWORD 3
 #define SSH_AUTH_ERROR 4
+#define SSH_CONN_UNKNOWN 1
+#define SSH_CONN_ESTABLISHED 2
 
-char *szBannerMsg = NULL;
+typedef struct __SSH2_DATA {
+  char *szBannerMsg;
+  int iConnectionStatus;
+} _SSH2_DATA;
 
-typedef struct __ssh2_data {
+typedef struct __ssh2_session_data {
   char *pPass;
   int iAnswerCount;
-} _ssh2_data;
-
+} _ssh2_session_data;
+  
 // Tells us whether we are to continue processing or not
 enum MODULE_STATE
 {
@@ -67,8 +72,8 @@ enum MODULE_STATE
 };
 
 // Forward declarations
-int tryLogin(int hSocket, LIBSSH2_SESSION *session, sLogin** login, char* szLogin, char* szPassword);
-int initModule(sLogin* login);
+int tryLogin(_SSH2_DATA* _psSessionData, int hSocket, LIBSSH2_SESSION *session, sLogin** login, char* szLogin, char* szPassword);
+int initModule(sLogin* login, _SSH2_DATA *_psSessionData);
 
 // Tell medusa how many parameters this module allows
 int getParamNumber()
@@ -110,6 +115,9 @@ int go(sLogin* logins, int argc, char *argv[])
 {
   int i;
   char *strtok_ptr, *pOpt, *pOptTmp;
+  _SSH2_DATA *psSessionData = NULL;
+  psSessionData = malloc(sizeof(_SSH2_DATA));
+  memset(psSessionData, 0, sizeof(_SSH2_DATA));
 
   if ( !(0 <= argc <= 1) )
   {
@@ -133,8 +141,8 @@ int go(sLogin* logins, int argc, char *argv[])
 
         if ( pOpt )
         {
-          szBannerMsg = malloc(strlen(pOpt));
-          strncpy((char *) szBannerMsg, pOpt, strlen(pOpt));
+          psSessionData->szBannerMsg = malloc(strlen(pOpt));
+          strncpy((char *) psSessionData->szBannerMsg, pOpt, strlen(pOpt));
         }
         else
         {
@@ -149,7 +157,7 @@ int go(sLogin* logins, int argc, char *argv[])
       free(pOptTmp);
     }
 
-    initModule(logins);
+    initModule(logins, psSessionData);
   }  
 
   return 0;
@@ -164,7 +172,8 @@ static LIBSSH2_ALLOC_FUNC(ssh2_init_alloc_cb)
 // void name(void *ptr, void **abstract)
 static LIBSSH2_FREE_FUNC(ssh2_init_free_cb)
 {
-  free(ptr);
+  /* Starting with libssh2 0.18 I'm getting a double free with this... */
+  //free(ptr);
 }
 
 // void *name(void *ptr, size_t count, void **abstract)
@@ -173,7 +182,7 @@ static LIBSSH2_REALLOC_FUNC(ssh2_init_realloc_cb)
   return realloc(ptr, count);
 }
 
-int initModule(sLogin* psLogin)
+int initModule(sLogin* psLogin, _SSH2_DATA *_psSessionData)
 {
   int hSocket = -1;
   enum MODULE_STATE nState = MSTATE_NEW;
@@ -187,9 +196,11 @@ int initModule(sLogin* psLogin)
   char *pErrorMsg;
   int iErrorMsg;
 
-  _ssh2_data ssh2_data;
-  ssh2_data.pPass = NULL;
-  ssh2_data.iAnswerCount = 0;
+  _psSessionData->iConnectionStatus = SSH_CONN_UNKNOWN;
+  
+  _ssh2_session_data ssh2_session_data;
+  ssh2_session_data.pPass = NULL;
+  ssh2_session_data.iAnswerCount = 0;
 
   memset(&params, 0, sizeof(sConnectParams));
  
@@ -230,7 +241,7 @@ int initModule(sLogin* psLogin)
           libssh2_session_free(session);
         }
         
-        session = libssh2_session_init_ex(ssh2_init_alloc_cb, ssh2_init_free_cb, ssh2_init_realloc_cb, &ssh2_data);
+        session = libssh2_session_init_ex(ssh2_init_alloc_cb, ssh2_init_free_cb, ssh2_init_realloc_cb, &ssh2_session_data);
         if (!session)
         {
           writeError(ERR_ERROR, "%s: Failed initiating SSH session: Host: %s User: %s Pass: %s", MODULE_NAME, psLogin->psServer->pHostIP, user->pUser, pPass);
@@ -241,24 +252,24 @@ int initModule(sLogin* psLogin)
 
         /* Set client SSH banner */
            
-        if (szBannerMsg) {
-          if ((strncmp(szBannerMsg, "SSH-2.0-", 8) != 0) || (strlen(szBannerMsg) > 253)) {
+        if (_psSessionData->szBannerMsg) {
+          if ((strncmp(_psSessionData->szBannerMsg, "SSH-2.0-", 8) != 0) || (strlen(_psSessionData->szBannerMsg) > 253)) {
             writeError(ERR_ERROR, "[%s] The selected banner could be rejected be some SSH servers. The banner should begin with \"SSH-2.0-\" and be shorter than 253 characters.", MODULE_NAME);
           } 
         } 
         else {
-          szBannerMsg = malloc(20);
-          memset(szBannerMsg, 0, 20);
-          sprintf(szBannerMsg, "SSH-2.0-MEDUSA_1.0");
+          _psSessionData->szBannerMsg = malloc(20);
+          memset(_psSessionData->szBannerMsg, 0, 20);
+          sprintf(_psSessionData->szBannerMsg, "SSH-2.0-MEDUSA_1.0");
         }
         
-        writeError(ERR_DEBUG_MODULE, "Attempting to set banner: %s", szBannerMsg);
-        if ( libssh2_banner_set(session, szBannerMsg) ) {
+        writeError(ERR_DEBUG_MODULE, "Attempting to set banner: %s", _psSessionData->szBannerMsg);
+        if ( libssh2_banner_set(session, _psSessionData->szBannerMsg) ) {
            writeError(ERR_DEBUG_MODULE, "Failed to set libssh banner.");
         }
         
         writeError(ERR_DEBUG_MODULE, "Attempting to initiate SSH session.");
-        for (i = 1; i <= 5; i++) {
+        for (i = 1; i <= psLogin->psServer->psHost->iRetries + 1; i++) {
           if (hSocket > 0) {
             medusaDisconnect(hSocket);
           }
@@ -266,28 +277,30 @@ int initModule(sLogin* psLogin)
           hSocket = medusaConnect(&params);
           if ( hSocket < 0 ) {
             writeError(ERR_NOTICE, "%s: failed to connect, port %d was not open on %s", MODULE_NAME, params.nPort, psLogin->psServer->pHostIP);
-            libssh2_session_free(session);
+            if (session)
+              libssh2_session_free(session);
             psLogin->iResult = LOGIN_RESULT_UNKNOWN;
             psLogin->iStatus = LOGIN_FAILED;
             return FAILURE;
           }
           
           if (libssh2_session_startup(session, hSocket)) {
-            writeError(ERR_ERROR, "%s: Failed establishing SSH session (%d/5): Host: %s User: %s Pass: %s", MODULE_NAME, i, psLogin->psServer->pHostIP, user->pUser, pPass);
+            writeError(ERR_ERROR, "%s: Failed establishing SSH session (%d/%d): Host: %s User: %s Pass: %s", MODULE_NAME, i, psLogin->psServer->psHost->iRetries + 1, psLogin->psServer->pHostIP, user->pUser, pPass);
           
             libssh2_session_last_error(session, &pErrorMsg, &iErrorMsg, 1);
             if ( (pErrorMsg) && (strstr(pErrorMsg, "Unable to exchange encryption keys")) ) {
               writeError(ERR_ERROR, "[%s] Failed to exchange encryption keys. Are you sure this is a SSHv2 server?", MODULE_NAME);
-              i = 5;
+              i = psLogin->psServer->psHost->iRetries + 1;
             }
             else if (pErrorMsg) {
               writeError(ERR_DEBUG_MODULE, "libssh2 Error Message: %s", pErrorMsg);
             }
 
-            if (i == 5) {
+            if (i == psLogin->psServer->psHost->iRetries + 1) {
               writeError(ERR_ERROR, "%s: Failed establishing SSH session. The following credentials were NOT tested: Host: %s User: %s Pass: %s", MODULE_NAME, psLogin->psServer->pHostIP, user->pUser, pPass);
               libssh2_session_disconnect(session, "");
-              libssh2_session_free(session);
+              if (session)
+                libssh2_session_free(session);
               if (hSocket > 0)
                 medusaDisconnect(hSocket);
               hSocket = -1;
@@ -305,9 +318,9 @@ int initModule(sLogin* psLogin)
         nState = MSTATE_RUNNING;
         break;
       case MSTATE_RUNNING:
-        ssh2_data.pPass = pPass;
-        ssh2_data.iAnswerCount = 0;
-        nState = tryLogin(hSocket, session, &psLogin, user->pUser, pPass);
+        ssh2_session_data.pPass = pPass;
+        ssh2_session_data.iAnswerCount = 0;
+        nState = tryLogin(_psSessionData, hSocket, session, &psLogin, user->pUser, pPass);
         if (psLogin->iResult != LOGIN_RESULT_UNKNOWN)
           pPass = getNextPass(psLogin->psServer->psAudit, user);
         break;
@@ -317,7 +330,8 @@ int initModule(sLogin* psLogin)
       default:
         writeError(ERR_CRITICAL, "Unknown %s module state %d", MODULE_NAME, nState);
         libssh2_session_disconnect(session, "");
-        libssh2_session_free(session);
+        if (session)
+          libssh2_session_free(session);
         if (hSocket > 0)
           medusaDisconnect(hSocket);
         hSocket = -1;
@@ -348,9 +362,9 @@ int initModule(sLogin* psLogin)
 void response_callback(const char* name, int name_len, const char* instruction, int instruction_len, int num_prompts, const LIBSSH2_USERAUTH_KBDINT_PROMPT* prompts, LIBSSH2_USERAUTH_KBDINT_RESPONSE* responses, void **abstract)
 {
     int i;
-    char *pPass = ((_ssh2_data*)(*abstract))->pPass;
+    char *pPass = ((_ssh2_session_data*)(*abstract))->pPass;
   
-    if (((_ssh2_data*)(*abstract))->iAnswerCount > 0)
+    if (((_ssh2_session_data*)(*abstract))->iAnswerCount > 0)
     {
       writeError(ERR_DEBUG_MODULE, "libssh2 response_callback: sshd asked a question, but we've already given out answer.");
     }
@@ -365,7 +379,7 @@ void response_callback(const char* name, int name_len, const char* instruction,
           strncpy(responses[i].text, pPass, strlen(pPass));
           responses[i].length = strlen(pPass);
           writeError(ERR_DEBUG_MODULE, "libssh2 response_callback set password response: %s", pPass);
-          ((_ssh2_data*)(*abstract))->iAnswerCount++;
+          ((_ssh2_session_data*)(*abstract))->iAnswerCount++;
         }
         else
         {
@@ -375,7 +389,7 @@ void response_callback(const char* name, int name_len, const char* instruction,
     }
 }
 
-int tryLogin(int hSocket, LIBSSH2_SESSION *session, sLogin** psLogin, char* szLogin, char* szPassword)
+int tryLogin(_SSH2_DATA* _psSessionData, int hSocket, LIBSSH2_SESSION *session, sLogin** psLogin, char* szLogin, char* szPassword)
 {
   char *pErrorMsg = NULL;
   int iErrorMsg, iAuthMode, iRet;
@@ -403,11 +417,13 @@ int tryLogin(int hSocket, LIBSSH2_SESSION *session, sLogin** psLogin, char* szLo
       if (strcmp(pAuth, "password") == 0) {
         writeError(ERR_DEBUG_MODULE, "Server support user-auth type: password");
         iAuthMode = SSH_AUTH_PASSWORD;
+        _psSessionData->iConnectionStatus = SSH_CONN_ESTABLISHED;
         break;
       }
       else if (strcmp(pAuth, "keyboard-interactive") == 0) {
         writeError(ERR_DEBUG_MODULE, "Server support user-auth type: keyboard-interactive");
         iAuthMode = SSH_AUTH_KBDINT;
+        _psSessionData->iConnectionStatus = SSH_CONN_ESTABLISHED;
         break;
       }
 
@@ -416,6 +432,13 @@ int tryLogin(int hSocket, LIBSSH2_SESSION *session, sLogin** psLogin, char* szLo
   
     free(pErrorMsg); 
   }
+  else if (_psSessionData->iConnectionStatus == SSH_CONN_ESTABLISHED)
+  {
+    writeError(ERR_DEBUG_MODULE, "Failed to retrieve supported authentication modes. Since previous connections worked, restarting entire session and attempting again.");
+    (*psLogin)->iResult = LOGIN_RESULT_UNKNOWN;
+    iRet = MSTATE_NEW;
+    return(iRet);
+  }
   else {
     writeError(ERR_ERROR, "Failed to retrieve supported authentication modes. Aborting...");
     (*psLogin)->iResult = LOGIN_RESULT_UNKNOWN;
