@@ -47,10 +47,16 @@
 #define  MODULE_NAME    "telnet.mod"
 #define MODULE_AUTHOR  "fizzgig <fizzgig@foofus.net>"
 #define  MODULE_SUMMARY_USAGE  "Brute force module for telnet sessions"
-#define MODULE_VERSION    "1.2.1"
-#define MODULE_VERSION_SVN "$Id: telnet.c 606 2006-10-19 20:22:41Z jmk $"
+#define MODULE_VERSION    "1.2.2"
+#define MODULE_VERSION_SVN "$Id: telnet.c 1040 2009-01-19 16:43:59Z jmk $"
 #define MODULE_SUMMARY_FORMAT  "%s : version %s"
 
+#define FREE(x) \
+        if (x != NULL) { \
+           free(x); \
+           x = NULL; \
+        }
+
 #define PORT_TELNET 23
 #define PORT_TELNETS 992
 
@@ -61,11 +67,18 @@
 #define RECEIVE_DELAY_1 20 * 1000000
 #define RECEIVE_DELAY_2 0.5 * 1000000
 
-const unsigned int BUFFER_SIZE = 300;
-const char* KNOWN_PROMPTS = ">#$%/";  // Each character represents a known telnet prompt - feel free to add a new one if desired
+#define MODE_NORMAL 0
+#define MODE_AS400 1
 
-const int KNOWN_PWD_SIZE = 3;  // Make sure to keep this in sync with the size of the array below!!
-const char* KNOWN_PWD_PROMPTS[] = { "assword", "asscode", "ennwort" };  // Complete/partial lines that indicate a password request
+typedef struct __MODULE_DATA {
+  int nMode;
+} _MODULE_DATA;
+
+const unsigned int BUFFER_SIZE = 300;
+const char* KNOWN_PROMPTS = ">#$%/?";  // Each character represents a known telnet prompt - feel free to add a new one if desired
+
+const int KNOWN_PWD_SIZE = 4;  // Make sure to keep this in sync with the size of the array below!!
+const char* KNOWN_PWD_PROMPTS[] = { "assword", "asscode", "ennwort", "ASSWORD" };  // Complete/partial lines that indicate a password request
 
 //const int KNOWN_LOGIN_SIZE = 4;  // Make sure to keep this in sync with the size of the array below!!
 //const char* KNOWN_LOGIN_PROMPTS[] = { "login:", "last login", "sername:", "User" }; // Complete/partial lines that request a user name
@@ -73,17 +86,19 @@ const int KNOWN_LOGIN_SIZE = 3;  // Make sure to keep this in sync with the size
 const char* KNOWN_LOGIN_PROMPTS[] = { "login:", "sername:", "User" }; // Complete/partial lines that request a user name
 
 // Tells us whether we are to continue processing or not
-enum TELNET_STATE
+enum MODULE_STATE
 {
-  TSTATE_NEW,
-  TSTATE_RUNNING,
-  TSTATE_EXITING
+  MSTATE_NEW,
+  MSTATE_RUNNING,
+  MSTATE_EXITING
 };
 
 // Forward declarations
-int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int nFoundPrompt);
-int initTelnet(sLogin* login);
-void processIAC(int hSocket, char** buffer, int* nBufferSize);
+int tryLogin(int hSocket, sLogin** login, _MODULE_DATA* _psSessionData, char* szLogin, char* szPassword, int nFoundPrompt);
+int tryLoginAS400(int hSocket, sLogin** login, _MODULE_DATA* _psSessionData, char* szLogin, char* szPassword);
+int initModule(sLogin* login, _MODULE_DATA* _psSessionData);
+int processIAClogout(int hSocket, _MODULE_DATA* _psSessionData);
+void processIAC(int hSocket, _MODULE_DATA* _psSessionData, char** buffer, int* nBufferSize);
 
 // Tell medusa how many parameters this module allows
 int getParamNumber()
@@ -114,12 +129,26 @@ void summaryUsage(char **ppszSummary)
 void showUsage()
 {
   writeVerbose(VB_NONE, "%s (%s) %s :: %s\n", MODULE_NAME, MODULE_VERSION, MODULE_AUTHOR, MODULE_SUMMARY_USAGE);
+  writeVerbose(VB_NONE, "Available module options:");
+  writeVerbose(VB_NONE, "  MODE:? (NORMAL, AS400) [optional]");
+  writeVerbose(VB_NONE, "    Sets the mode for error detection.");
+  writeVerbose(VB_NONE, "");
+  writeVerbose(VB_NONE, "  Usage example: \"-M telnet -m MODE:AS400 -U accounts.txt -p password\"");
+  writeVerbose(VB_NONE, "");
 }
 
 // The "main" of the medusa module world - this is what gets called to actually do the work
 int go(sLogin* logins, int argc, char *argv[])
 {
-  if (argc != 0)
+  int i;
+
+  char *strtok_ptr, *pOpt, *pOptTmp;
+  _MODULE_DATA *psSessionData;
+
+  psSessionData = malloc(sizeof(_MODULE_DATA));
+  memset(psSessionData, 0, sizeof(_MODULE_DATA));
+
+  if ( !(0 <= argc <= 3) )
   {
     // Show usage information
     writeError(ERR_ERROR, "%s is expecting 0 parameters, but it was passed %d", MODULE_NAME, argc);
@@ -128,16 +157,41 @@ int go(sLogin* logins, int argc, char *argv[])
   else
   {
     writeError(ERR_DEBUG_MODULE, "OMG teh %s module has been called!!", MODULE_NAME);
-    initTelnet(logins);
+
+    for (i=0; i<argc; i++) {
+      pOptTmp = malloc( strlen(argv[i]) + 1);
+      memset(pOptTmp, 0, strlen(argv[i]) + 1);
+      strncpy(pOptTmp, argv[i], strlen(argv[i]));
+      writeError(ERR_DEBUG_MODULE, "Processing complete option: %s", pOptTmp);
+      pOpt = strtok_r(pOptTmp, ":", &strtok_ptr);
+      writeError(ERR_DEBUG_MODULE, "Processing option: %s", pOpt);
+
+      if (strcmp(pOpt, "MODE") == 0)
+      {
+        pOpt = strtok_r(NULL, "\0", &strtok_ptr);
+        writeError(ERR_DEBUG_MODULE, "Processing option parameter: %s", pOpt);
+
+        if (strcmp(pOpt, "AS400") == 0)
+          psSessionData->nMode = MODE_AS400;
+        else
+          writeError(ERR_WARNING, "Invalid value for method MODE.");
+      }
+      else
+         writeError(ERR_WARNING, "Invalid method: %s.", pOpt);
+
+      free(pOptTmp);
+    }
+
+    initModule(logins, psSessionData);
   }
 
   return SUCCESS;
 }
 
-int initTelnet(sLogin* login)
+int initModule(sLogin* login, _MODULE_DATA *_psSessionData)
 {
   int hSocket = -1;
-  enum TELNET_STATE nState = TSTATE_NEW;
+  enum MODULE_STATE nState = MSTATE_NEW;
   char* bufReceive;
   int nReceiveBufferSize = 0, nFoundPrompt = PROMPT_UNKNOWN;
   int i = 0;
@@ -174,21 +228,21 @@ int initTelnet(sLogin* login)
   {
     switch(nState)
     {
-    case TSTATE_NEW:
+    case MSTATE_NEW:
       // Already have an open socket - close it
       if (hSocket > 0)
         medusaDisconnect(hSocket);
 
       if (login->psServer->psHost->iUseSSL > 0)
         hSocket = medusaConnectSSL(&params);
-      else               
+      else 
         hSocket = medusaConnect(&params);
       
       if (hSocket <= 0)
       {
-        writeError(ERR_NOTICE, "%s: failed to connect, port %d was not open on %s", MODULE_NAME, params.nPort, login->psServer->pHostIP);
+        writeError(ERR_ERROR, "[%s] Failed to connect, port %d was not open on %s", MODULE_NAME, params.nPort, login->psServer->pHostIP);
         login->iResult = LOGIN_RESULT_UNKNOWN;
-        login->iStatus = LOGIN_DONE;
+        login->iStatus = LOGIN_FAILED;
         setPassResult(login, pPass);
         return FAILURE;
       }
@@ -205,9 +259,9 @@ int initTelnet(sLogin* login)
 
       if (bufReceive == NULL)
       {
-        writeError(ERR_NOTICE, "null response was unexpected from a telnet server (is one running?)");
+        writeError(ERR_ERROR, "[%s] null response was unexpected from a telnet server (is one running?)", MODULE_NAME);
         login->iResult = LOGIN_RESULT_UNKNOWN;
-        login->iStatus = LOGIN_DONE;
+        login->iStatus = LOGIN_FAILED;
         setPassResult(login, pPass);
         return FAILURE;
       }
@@ -216,7 +270,7 @@ int initTelnet(sLogin* login)
       do
       {
         nFoundPrompt = PROMPT_UNKNOWN;
-        processIAC(hSocket, &bufReceive, &nReceiveBufferSize);
+        processIAC(hSocket, _psSessionData, &bufReceive, &nReceiveBufferSize);
 
         if (bufReceive != NULL && bufReceive[0] != 0 && (unsigned char)bufReceive[0] != IAC)
           makeToLower(bufReceive);
@@ -224,81 +278,140 @@ int initTelnet(sLogin* login)
         if (bufReceive != NULL)
         {
           writeError(ERR_DEBUG_MODULE, "Looking for login prompts");
-          
-          // Look for known login prompts
-          for (i = 0; i < KNOWN_LOGIN_SIZE; i++)
+
+          if (_psSessionData->nMode == MODE_AS400)
           {
-            if (strstr(bufReceive, KNOWN_LOGIN_PROMPTS[i]) != NULL)
+            if (strcasestr(bufReceive, "Sign On") != NULL)
             {
-              // Do we have a prompt?
-              writeError(ERR_DEBUG_MODULE, "Found login prompt...");
+              writeError(ERR_INFO, "[%s] Detected AS/400 Sign On Screen.", MODULE_NAME);
               nFoundPrompt = PROMPT_LOGIN_PASSWORD;
+              
+              FREE(bufReceive);
+              if (medusaDataReadyTimed(hSocket, 0, 20000) > 0)
+              {
+                // More data waiting
+                bufReceive = medusaReceiveLineDelay(hSocket, &nReceiveBufferSize, RECEIVE_DELAY_1, RECEIVE_DELAY_2);
+                if (bufReceive != NULL)
+                  bufReceive[nReceiveBufferSize] = 0;  // Make certain buffer is null-terminated
+              }
+ 
               break;
             }
+
+            /*
+            Sign On
+            System  . . . . . :   TSTDBS16
+            Subsystem . . . . :   QINTER
+            Display . . . . . :   QPADEV0001
+            */
           }
-          
-          /* Some systems do not provide a login prompt and go right to password */
-          for (i = 0; i < KNOWN_PWD_SIZE; i++)
+          else
           {
-            if (strstr(bufReceive, KNOWN_PWD_PROMPTS[i]) != NULL)
+            // Look for known login prompts
+            for (i = 0; i < KNOWN_LOGIN_SIZE; i++)
             {
-              // Do we have a prompt?
-              writeError(ERR_DEBUG_MODULE, "Found a password prompt already...");
-              nFoundPrompt = PROMPT_PASSWORD;
-              break;
+              if (strcasestr(bufReceive, KNOWN_LOGIN_PROMPTS[i]) != NULL)
+              {
+                // Do we have a prompt?
+                writeError(ERR_DEBUG_MODULE, "Found login prompt...");
+                nFoundPrompt = PROMPT_LOGIN_PASSWORD;
+                break;
+              }
             }
-          }
           
-          if (nFoundPrompt == PROMPT_UNKNOWN)
-          {
-            if (medusaDataReadyTimed(hSocket, 0, 20000) > 0)
+            /* Some systems do not provide a login prompt and go right to password */
+            for (i = 0; i < KNOWN_PWD_SIZE; i++)
             {
-              // More data waiting
-              bufReceive = medusaReceiveLineDelay(hSocket, &nReceiveBufferSize, RECEIVE_DELAY_1, RECEIVE_DELAY_2);
-              if (bufReceive != NULL)
-                bufReceive[nReceiveBufferSize] = 0;  // Make certain buffer is null-terminated
-            } 
+              if (strcasestr(bufReceive, KNOWN_PWD_PROMPTS[i]) != NULL)
+              {
+                // Do we have a prompt?
+                writeError(ERR_DEBUG_MODULE, "Found a password prompt already...");
+                nFoundPrompt = PROMPT_PASSWORD;
+                
+                if (login->psServer->iLoginsDone < 1 && login->iId == 0)
+                  writeVerbose(VB_NONE_FILE, "Password Prompt Only: %s\n", login->psServer->pHostIP);
+                
+                break;
+              }
+            }
+          
+            if (nFoundPrompt == PROMPT_UNKNOWN)
+            {
+              FREE(bufReceive);
+              if (medusaDataReadyTimed(hSocket, 0, 20000) > 0)
+              {
+                // More data waiting
+                bufReceive = medusaReceiveLineDelay(hSocket, &nReceiveBufferSize, RECEIVE_DELAY_1, RECEIVE_DELAY_2);
+                if (bufReceive != NULL)
+                  bufReceive[nReceiveBufferSize] = 0;  // Make certain buffer is null-terminated
+              } 
+            }
           }
         }
       }
       while (bufReceive != NULL && (unsigned char)bufReceive[0] == IAC && nFoundPrompt == PROMPT_UNKNOWN);
 
-      if (bufReceive != NULL)
-        free(bufReceive);
+      FREE(bufReceive);
 
-      nState = TSTATE_RUNNING;
+      if (nFoundPrompt == PROMPT_UNKNOWN)
+      {
+        writeError(ERR_ERROR, "[%s] Failed to identify logon prompt.", MODULE_NAME); 
+        login->iResult = LOGIN_RESULT_UNKNOWN;
+        login->iStatus = LOGIN_FAILED;
+        setPassResult(login, pPass);
+        return FAILURE;
+      }
+      else
+        nState = MSTATE_RUNNING;
+      
       break;
 
-    case TSTATE_RUNNING:
-      nState = tryLogin(hSocket, &login, user->pUser, pPass, nFoundPrompt);
-      if (nState == TSTATE_EXITING)
+    case MSTATE_RUNNING:
+      if (_psSessionData->nMode == MODE_AS400)
+        nState = tryLoginAS400(hSocket, &login, _psSessionData, user->pUser, pPass);
+      else
+        nState = tryLogin(hSocket, &login, _psSessionData, user->pUser, pPass, nFoundPrompt);
+      
+      if (nState == MSTATE_EXITING)
       {
         pPass = NULL;
       }
       else
       {
+        if (processIAClogout(hSocket, _psSessionData) == FAILURE)
+        {
+          writeError(ERR_ERROR, "[%s] Failed to close existing Telnet session.", MODULE_NAME);
+        }
         medusaDisconnect(hSocket);
         hSocket = -1;
+        
+        /*
+          Cisco devices appear to keep sessions open for a brief time after we terminate 
+          the connection. They also seem to ignore "IAC DO LOGOUT" commands. Adding a 
+          sleep() hack here, to give them some time to clean-up. 
+        */
+        sleep(1);
+
         pPass = getNextPass(login->psServer->psAudit, user);
         if (pPass == NULL)
         {
           writeVerbose(VB_GENERAL, "Done with passwords, exiting telnet");
-          nState = TSTATE_EXITING;
+          nState = MSTATE_EXITING;
         }
         else
         {
-          nState = TSTATE_NEW;
+          nState = MSTATE_NEW;
         }
       }
       break;
-    case TSTATE_EXITING:
+    case MSTATE_EXITING:
       if (hSocket > 0)
         medusaDisconnect(hSocket);
       hSocket = -1;
     default:
-      writeError(ERR_CRITICAL, "unknown telnet module state");
+      writeError(ERR_CRITICAL, "Unknown %s module (%d) state %d host: %s", MODULE_NAME, login->iId, nState, login->psServer->pHostIP);
       login->iResult = LOGIN_RESULT_UNKNOWN;
-      login->iStatus = LOGIN_DONE;
+      login->iStatus = LOGIN_FAILED;
       setPassResult(login, pPass);
     }
   }
@@ -307,9 +420,9 @@ int initTelnet(sLogin* login)
   return SUCCESS;
 }
 
-int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int nFoundPrompt)
+int tryLogin(int hSocket, sLogin** login, _MODULE_DATA* _psSessionData, char* szLogin, char* szPassword, int nFoundPrompt)
 {
-  // This function should return TSTATE_RUNNING to continue or TSTATE_EXITING to terminate the module
+  // This function should return MSTATE_RUNNING to continue or MSTATE_EXITING to terminate the module
   char bufSend[BUFFER_SIZE];
   char* bufReceive;
   int nSendBufferSize = 0, nReceiveBufferSize = 0;
@@ -321,7 +434,7 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
     writeError(ERR_ERROR, "%s failed: socket was invalid", MODULE_NAME);
     (*login)->iResult = LOGIN_RESULT_UNKNOWN;
     setPassResult(*login, szPassword);
-    return TSTATE_EXITING;    // No good socket
+    return MSTATE_EXITING;    // No good socket
   }
 
   if (nFoundPrompt == PROMPT_LOGIN_PASSWORD)
@@ -336,7 +449,7 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
       writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
       (*login)->iResult = LOGIN_RESULT_UNKNOWN;
       setPassResult(*login, szPassword);
-      return TSTATE_EXITING;
+      return MSTATE_EXITING;
     }
 
     do
@@ -349,7 +462,7 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
         writeError(ERR_ERROR, "%s: Telnet did not respond to the sending of the user name '%s' in a timely fashion - is it down or refusing connections?", MODULE_NAME, szLogin);
         (*login)->iResult = LOGIN_RESULT_UNKNOWN;
         setPassResult(*login, szPassword);
-        return TSTATE_EXITING;
+        return MSTATE_EXITING;
       }
 
       bufReceive[nReceiveBufferSize] = 0;  // Make certain buffer is null-terminated
@@ -360,7 +473,7 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
         (*login)->iResult = LOGIN_RESULT_SUCCESS;
         setPassResult(*login, szPassword);
         free(bufReceive);
-        return TSTATE_EXITING;
+        return MSTATE_EXITING;
       }
 
       makeToLower(bufReceive);
@@ -368,7 +481,7 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
       // Are we at a known password prompt?
       for (i = 0; i < KNOWN_PWD_SIZE; i++)
       {
-        if (strstr(bufReceive, KNOWN_PWD_PROMPTS[i]) != NULL)
+        if (strcasestr(bufReceive, KNOWN_PWD_PROMPTS[i]) != NULL)
         {
           nContinue = 1;
           break;
@@ -380,12 +493,12 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
       {
         for (i = 0; i < KNOWN_LOGIN_SIZE; i++)
         {
-          if (strstr(bufReceive, KNOWN_LOGIN_PROMPTS[i]) != NULL)
+          if (strcasestr(bufReceive, KNOWN_LOGIN_PROMPTS[i]) != NULL)
           {
             free(bufReceive);
             (*login)->iResult = LOGIN_RESULT_FAIL;
             setPassResult(*login, szPassword);
-            return TSTATE_RUNNING;
+            return MSTATE_RUNNING;
           }
         }
       }
@@ -396,11 +509,11 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
   }
   else if (nFoundPrompt == PROMPT_PASSWORD)
   {
-    writeError(ERR_DEBUG_MODULE, "%s we are skipping a username", MODULE_NAME);
+    writeError(ERR_DEBUG_MODULE, "[%s] we are skipping a username", MODULE_NAME);
   }
   else
   {
-    writeError(ERR_ERROR, "%s: No login prompt detected.", MODULE_NAME);
+    writeError(ERR_ERROR, "[%s] No login prompt detected.", MODULE_NAME);
     return FAILURE;
   }
 
@@ -414,7 +527,7 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
     writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
     (*login)->iResult = LOGIN_RESULT_UNKNOWN;
     setPassResult(*login, szPassword);
-    return TSTATE_EXITING;
+    return MSTATE_EXITING;
   }
 
   // Look for a return
@@ -424,14 +537,14 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
     writeError(ERR_ERROR, "timeout waiting for response from server after sending password");
     (*login)->iResult = LOGIN_RESULT_UNKNOWN;
     setPassResult(*login, szPassword);
-    return TSTATE_EXITING;
+    return MSTATE_EXITING;
   }
 
   bufReceive[nReceiveBufferSize] = 0;  // Make certain buffer is null-terminated
 
   // It's possible that some telnet servers (like Microsoft's) may send some more IAC commands at this point
   // Take care of zem!
-  processIAC(hSocket, &bufReceive, &nReceiveBufferSize);
+  processIAC(hSocket, _psSessionData, &bufReceive, &nReceiveBufferSize);
 
   if (bufReceive == NULL)
     bufReceive = medusaReceiveLineDelay(hSocket, &nReceiveBufferSize, RECEIVE_DELAY_1, RECEIVE_DELAY_2);
@@ -440,12 +553,19 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
   while (bufReceive != NULL)
   {
     /* check for known failures */
-    if (strstr(bufReceive, "% Authentication failed."))
+    if (strstr(bufReceive, "Authentication failed"))
     {
       writeError(ERR_DEBUG, "Server responded with Cisco \"Authentication failed.\" message.");
       (*login)->iResult = LOGIN_RESULT_FAIL;
       setPassResult(*login, szPassword);
-      return TSTATE_NEW;
+      return MSTATE_NEW;
+    }
+    if (strstr(bufReceive, "Login invalid"))
+    {
+      writeError(ERR_DEBUG, "Server responded with Cisco \"Login invalid\" message.");
+      (*login)->iResult = LOGIN_RESULT_FAIL;
+      setPassResult(*login, szPassword);
+      return MSTATE_NEW;
     }
     else if (strcspn(bufReceive, KNOWN_PROMPTS) != strlen(bufReceive))
     {
@@ -453,7 +573,7 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
       free(bufReceive);
       (*login)->iResult = LOGIN_RESULT_SUCCESS;
       setPassResult(*login, szPassword);
-      return TSTATE_EXITING;
+      return MSTATE_EXITING;
     }
     else
     {
@@ -461,13 +581,13 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
         // If we have a login prompt, we have failed
         for (i = 0; i < KNOWN_LOGIN_SIZE; i++)
         {
-          if (strstr(bufReceive, KNOWN_LOGIN_PROMPTS[i]) != NULL)
+          if (strcasestr(bufReceive, KNOWN_LOGIN_PROMPTS[i]) != NULL)
           {
             free(bufReceive);
             writeError(ERR_DEBUG_MODULE, "unsuccessful login - user '%s' with a password of '%s'", szLogin, szPassword);
             (*login)->iResult = LOGIN_RESULT_FAIL;
             setPassResult(*login, szPassword);
-            return TSTATE_NEW;
+            return MSTATE_NEW;
           }
         }
       } 
@@ -476,13 +596,13 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
         // We are operating on a non-login telnet setup
         for (i = 0; i < KNOWN_PWD_SIZE; i++)
         {
-          if (strstr(bufReceive, KNOWN_PWD_PROMPTS[i]) != NULL)
+          if (strcasestr(bufReceive, KNOWN_PWD_PROMPTS[i]) != NULL)
           {
             free(bufReceive);
             writeError(ERR_DEBUG_MODULE, "unsuccessful login with a password of '%s'", szPassword);
             (*login)->iResult = LOGIN_RESULT_FAIL;
             setPassResult(*login, szPassword);
-            return TSTATE_NEW;
+            return MSTATE_NEW;
           }
         }
       }
@@ -494,32 +614,240 @@ int tryLogin(int hSocket, sLogin** login, char* szLogin, char* szPassword, int n
 
   (*login)->iResult = LOGIN_RESULT_FAIL;
   setPassResult(*login, szPassword);
-  return TSTATE_NEW;
+
+  return MSTATE_NEW;
 }
 
-void processIAC(int hSocket, char** buffer, int* nReceiveBufferSize)
+int tryLoginAS400(int hSocket, sLogin** psLogin, _MODULE_DATA* _psSessionData, char* szLogin, char* szPassword)
+{
+  char bufSend[BUFFER_SIZE];
+  char* bufReceive;
+  int nSendBufferSize = 0, nReceiveBufferSize = 0;
+  int iRet = FAILURE;
+  char szUser[10 + 1];
+  char szPass[128 + 1];
+  char szErrorMsg[100];
+
+  if (hSocket <= 0)
+  {
+    writeError(ERR_ERROR, "%s failed: socket was invalid", MODULE_NAME);
+    (*psLogin)->iResult = LOGIN_RESULT_UNKNOWN;
+    setPassResult(*psLogin, szPassword);
+    return MSTATE_EXITING;
+  }
+
+  /* Send username and password */
+  /* USERNAME \t (0x09) PASSWORD \r (0x0D) \0 (0x00) */
+
+  /* Password Policy Information
+    http://publib.boulder.ibm.com/iseries/v5r1/ic2924/index.htm?info/rzakz/rzakzqpwdlvl.htm
+
+    Short passwords: The AS/400 "short" passwords are 0-10 characters in length. They
+    allow the following characters: A-Z 0-9 $ @ # _
+  
+    Long passwords: The AS/400 "long" passwords are 0-128 characters in length. Upper and
+    lower case passwords consisting of any characters are allowed.
+
+    Usernames appear to be limited to 10 characters in length and use upper-case. 
+    ** This has not been fully verified. **
+
+    http://download.oracle.com/docs/html/B10256_01/ch2.htm
+    IBM DB2/400 V4R5 object names can be up to 10 alphanumeric characters in length, 
+    beginning with a letter or a national character.
+  */
+
+  memset(bufSend, 0, BUFFER_SIZE);
+  memset(szUser, 0, 10 + 1);
+  memset(szPass, 0, 128 + 1);
+  
+  strncpy(szUser, szLogin, 10);
+  strncpy(szPass, szPassword, 128);
+
+  sprintf(bufSend, "%s\t%s\r", szUser, szPass);
+  nSendBufferSize = strlen(bufSend) + 1;
+
+  if (medusaSend(hSocket, bufSend, nSendBufferSize, 0) < 0)
+  {
+    writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
+    (*psLogin)->iResult = LOGIN_RESULT_UNKNOWN;
+    setPassResult(*psLogin, szPassword);
+    return MSTATE_EXITING;
+  }
+
+  /* Process server response */
+  bufReceive = medusaReceiveLineDelay(hSocket, &nReceiveBufferSize, RECEIVE_DELAY_1, RECEIVE_DELAY_2);
+  if (bufReceive == NULL)
+  {
+    writeError(ERR_ERROR, "[%s] Timeout waiting for response from server after sending password", MODULE_NAME);
+    (*psLogin)->iResult = LOGIN_RESULT_UNKNOWN;
+    setPassResult(*psLogin, szPassword);
+    return MSTATE_EXITING;
+  }
+
+  if (strstr(bufReceive, "CPF1120") != NULL)
+  {
+    sprintf(szErrorMsg, "CPF1120 - User %s does not exist.", szUser); 
+    writeError(ERR_ERROR, "[%s] %s", MODULE_NAME, szErrorMsg);
+    (*psLogin)->pErrorMsg = malloc( strlen(szErrorMsg) + 1 );
+    memset((*psLogin)->pErrorMsg, 0, strlen(szErrorMsg) + 1 );
+    strncpy((*psLogin)->pErrorMsg, szErrorMsg, strlen(szErrorMsg));
+    (*psLogin)->iResult = LOGIN_RESULT_ERROR;
+    iRet = MSTATE_EXITING;
+  }
+  else if (strstr(bufReceive, "CPF1116") != NULL)
+  {
+    strcpy(szErrorMsg, "CPF1116 - Next not valid sign-on attempt varies off device."); 
+    writeError(ERR_ERROR, "[%s] %s", MODULE_NAME, szErrorMsg);
+    (*psLogin)->pErrorMsg = malloc( strlen(szErrorMsg) + 1 );
+    memset((*psLogin)->pErrorMsg, 0, strlen(szErrorMsg) + 1 );
+    strncpy((*psLogin)->pErrorMsg, szErrorMsg, strlen(szErrorMsg));
+    (*psLogin)->iResult = LOGIN_RESULT_FAIL;
+    iRet = MSTATE_NEW;
+  }
+  else if (strstr(bufReceive, "CPF1392") != NULL)
+  {
+    strcpy(szErrorMsg, "CPF1392 - Next not valid sign-on disables user profile."); 
+    writeError(ERR_ERROR, "[%s] %s", MODULE_NAME, szErrorMsg);
+    (*psLogin)->pErrorMsg = malloc( strlen(szErrorMsg) + 1 );
+    memset((*psLogin)->pErrorMsg, 0, strlen(szErrorMsg) + 1 );
+    strncpy((*psLogin)->pErrorMsg, szErrorMsg, strlen(szErrorMsg));
+    (*psLogin)->iResult = LOGIN_RESULT_ERROR;
+    iRet = MSTATE_EXITING;
+  }
+  /*
+  http://archive.midrange.com/midrange-l/200507/msg01092.html
+  
+  Cause . . . . . :   User profile &1 has reached the maximum number of
+                      sign-on attempts and has been disabled, or the STATUS 
+                      parameter has been changed to *DISABLED on the Create 
+                      User Profile (CRTUSRPRF) or Change User Profile
+                      (CHGUSRPRF) command.
+  
+  Recovery  . . . :   To enable the user profile, have the security officer
+                      change the STATUS parameter to *ENABLED on the Change 
+                      User Profile (CHGUSRPRF) command.
+  */
+  else if (strstr(bufReceive, "CPF1394") != NULL)
+  {
+    sprintf(szErrorMsg, "CPF1394 - User profile %s cannot sign on.", szUser); 
+    writeError(ERR_ERROR, "[%s] %s", MODULE_NAME, szErrorMsg);
+    (*psLogin)->pErrorMsg = malloc( strlen(szErrorMsg) + 1 );
+    memset((*psLogin)->pErrorMsg, 0, strlen(szErrorMsg) + 1 );
+    strncpy((*psLogin)->pErrorMsg, szErrorMsg, strlen(szErrorMsg));
+    (*psLogin)->iResult = LOGIN_RESULT_ERROR;
+    iRet = MSTATE_EXITING;
+  }
+  else if (strstr(bufReceive, "CPF1118") != NULL)
+  {
+    sprintf(szErrorMsg, "CPF1118 - No password associated with user %s.", szUser); 
+    writeError(ERR_ERROR, "[%s] %s", MODULE_NAME, szErrorMsg);
+    (*psLogin)->pErrorMsg = malloc( strlen(szErrorMsg) + 1 );
+    memset((*psLogin)->pErrorMsg, 0, strlen(szErrorMsg) + 1 );
+    strncpy((*psLogin)->pErrorMsg, szErrorMsg, strlen(szErrorMsg));
+    (*psLogin)->iResult = LOGIN_RESULT_SUCCESS;
+    iRet = MSTATE_EXITING;
+  }
+  else if (strstr(bufReceive, "CPF1109") != NULL)
+  {
+    strcpy(szErrorMsg, "CPF1109 - Not authorized to subsystem."); 
+    writeError(ERR_ERROR, "[%s] %s", MODULE_NAME, szErrorMsg);
+    (*psLogin)->pErrorMsg = malloc( strlen(szErrorMsg) + 1 );
+    memset((*psLogin)->pErrorMsg, 0, strlen(szErrorMsg) + 1 );
+    strncpy((*psLogin)->pErrorMsg, szErrorMsg, strlen(szErrorMsg));
+    (*psLogin)->iResult = LOGIN_RESULT_SUCCESS;
+    iRet = MSTATE_EXITING;
+  }
+  else if (strstr(bufReceive, "CPF1110") != NULL)
+  {
+    strcpy(szErrorMsg, "CPF1110 - Not authorized to work station."); 
+    writeError(ERR_ERROR, "[%s] %s", MODULE_NAME, szErrorMsg);
+    (*psLogin)->pErrorMsg = malloc( strlen(szErrorMsg) + 1 );
+    memset((*psLogin)->pErrorMsg, 0, strlen(szErrorMsg) + 1 );
+    strncpy((*psLogin)->pErrorMsg, szErrorMsg, strlen(szErrorMsg));
+    (*psLogin)->iResult = LOGIN_RESULT_SUCCESS;
+    iRet = MSTATE_EXITING;
+  }
+  else if (strstr(bufReceive, "CPF1107") != NULL)
+  {
+    writeError(ERR_INFO, "[%s] CPF1107 - Password not correct for user profile.", MODULE_NAME);
+    (*psLogin)->iResult = LOGIN_RESULT_FAIL;
+    iRet = MSTATE_NEW;
+  }
+  else
+  {
+    (*psLogin)->iResult = LOGIN_RESULT_SUCCESS;
+    iRet = MSTATE_EXITING;
+  }
+  
+  setPassResult((*psLogin), szPass);
+
+  return iRet;
+}
+
+/*
+   The sender of this command REQUESTS that the receiver forcibly log
+   off the user process at the receiver's end, or confirms that the
+   receiver has its permission to do so.
+*/
+int processIAClogout(int hSocket, _MODULE_DATA* _psSessionData)
+{
+  unsigned char bufSend[] = { 0xFF, 0xFD, 0x12 }; /* IAC DO LOGOUT */
+  char* bufReceive = NULL;
+  int nReceiveBufferSize = 0;
+
+  writeError(ERR_DEBUG_MODULE, "[%s] Sending IAC DO LOGOUT command.", MODULE_NAME);
+  if (medusaSend(hSocket, bufSend, 3, 0) < 0)
+  {
+    writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
+    return FAILURE;
+  }
+
+  /* Receive any remaining IAC commands */
+  /*
+  bufReceive = medusaReceiveLineDelay(hSocket, &nReceiveBufferSize, RECEIVE_DELAY_1, RECEIVE_DELAY_2);
+  if (bufReceive == NULL)
+    return FAILURE;
+
+  processIAC(hSocket, _psSessionData, &bufReceive, &nReceiveBufferSize);
+  */
+
+  return SUCCESS;
+}
+
+void processIAC(int hSocket, _MODULE_DATA* _psSessionData, char** buffer, int* nReceiveBufferSize)
 {
   unsigned char* bufTemp = (unsigned char*)*buffer;
 
-  while (*bufTemp == IAC)
+  /* We're not that friendly. Refuse to do anything asked of us. */
+  while (*bufTemp == IAC) /* IAC (0xFF) */
   {
-    writeError(ERR_DEBUG_MODULE, "Handling command...");
+    writeError(ERR_DEBUG_MODULE, "Handling IAC Command...");
 
     if ((bufTemp[1] == 0xfc || bufTemp[1] == 0xfe) && bufTemp[2] == 0x22)
     {
       writeError(ERR_DEBUG_MODULE, "TELNETD peer does not like linemode");
     }
-    if (bufTemp[1] == WILL || bufTemp[1] == WONT)
+
+    if (bufTemp[1] == WILL) /* WILL (0xFB), WONT (0xFC) */
     {
-      bufTemp[1] = DONT;
+      /* AS/400 devices appear to request and require "Echo" and "Suppress Go Ahead" */
+      if (_psSessionData->nMode == MODE_AS400)
+        if ((bufTemp[2] == TELOPT_ECHO) || (bufTemp[2] == TELOPT_SGA))
+          bufTemp[1] = DO;
+        else
+          bufTemp[1] = DONT;
+      else
+        bufTemp[1] = DONT;
+
       medusaSend(hSocket, bufTemp, 3, 0);
     }
-    else if (bufTemp[1] == DO || bufTemp[1] == DONT)
+    else if (bufTemp[1] == DO) /* DO (0xFD), DONT (0xFE) */
     {
       bufTemp[1] = WONT;
       medusaSend(hSocket, bufTemp, 3, 0);
     }
-    bufTemp += 3; // Read three bytes - move on
+
+    bufTemp += 3; /* Process three bytes at a time */
   }
 
   if (*bufTemp == 0)
