@@ -30,12 +30,13 @@
  ***************************************************************************/
 
 #include "module.h"
+#include "math.h"
 
 #define MODULE_NAME    "web-form.mod"
 #define MODULE_AUTHOR  "Luciano Bello <luciano@linux.org.ar>"
 #define MODULE_SUMMARY_USAGE  "Brute force module for web forms"
-#define MODULE_VERSION    "0.9"
-#define MODULE_VERSION_SVN "$Id: web-form.c 0 2007-07-25 22:22:41Z luciano $"
+#define MODULE_VERSION    "1.0.0"
+#define MODULE_VERSION_SVN "$Id: web-form.c 980 2008-10-01 20:22:51Z jmk $"
 #define MODULE_SUMMARY_FORMAT  "%s : version %s"
 #define MODULE_SUMMARY_FORMAT_WARN  "%s : version %s (%s)"
 #define OPENSSL_WARNING "No usable OPENSSL. Module disabled."
@@ -48,8 +49,6 @@
 
 #ifdef HAVE_LIBSSL
 
-#include "../http-ntlm.h"
-
 #define HTTP_PORT 80
 #define HTTPS_PORT 443
 
@@ -119,7 +118,8 @@ void showUsage()
   writeVerbose(VB_NONE, "  FORM-DATA:<METHOD>?<FIELDS>");
   writeVerbose(VB_NONE, "                     Methods and fields to send to web service. Valid methods are GET and POST. The actual form");
   writeVerbose(VB_NONE, "                     data to be submitted should also be defined here. Specifically, the fields: username and");
-  writeVerbose(VB_NONE, "                     password. Default: \"post?username=&password=\"");
+  writeVerbose(VB_NONE, "                     password. The username field must be the first, followed by the password field.");
+  writeVerbose(VB_NONE, "                     Default: \"post?username=&password=\"");
   writeVerbose(VB_NONE, "");
   writeVerbose(VB_NONE, "Usage example: \"-M web-form -m USER-AGENT:\"g3rg3 gerg\" -m FORM:\"webmail/index.php\" -m DENY-SIGNAL:\"deny!\"");
   writeVerbose(VB_NONE, "                 -m FORM-DATA:\"post?user=&pass=&submit=True\"");
@@ -229,6 +229,7 @@ int initHttp(_MODULE_DATA *_psSessionData, sLogin* login)
   char* bufReceive;
   int i = 0;
   char *pPass;
+  char *pStrtokSavePtr = NULL;
   char *pTemp;
   int nPort, nBufLength = 0;
   sUser* user = login->psUser;
@@ -310,19 +311,42 @@ int initHttp(_MODULE_DATA *_psSessionData, sLogin* login)
           _psSessionData->nFormType = FORM_POST;
         }
         else {
-          pTemp = strtok(_psSessionData->szFormData, "?");
-          
-          writeError(ERR_DEBUG_MODULE, "[%s] User-supplied Form Action Method: %s", MODULE_NAME, pTemp);
-          if(strncasecmp(pTemp, "POST", 4) == 0) 
-            _psSessionData->nFormType=FORM_POST;
-          else if(strncasecmp(pTemp, "GET", 3) == 0) 
-            _psSessionData->nFormType=FORM_GET;
-          else 
-            _psSessionData->nFormType=FORM_UNKNOWN;
- 
-          _psSessionData->szFormUser = strtok(NULL, "&");
-          _psSessionData->szFormPass = strtok(NULL, "&");
-          _psSessionData->szFormRest = strtok(NULL, "");
+          /* Only set user-supplied form data on first pass */  
+          if (_psSessionData->szFormUser == NULL)
+          {
+            pTemp = strtok_r(_psSessionData->szFormData, "?", &pStrtokSavePtr);
+            writeError(ERR_DEBUG_MODULE, "[%s] User-supplied Form Action Method: %s", MODULE_NAME, pTemp);
+            if(strncasecmp(pTemp, "POST", 4) == 0) 
+              _psSessionData->nFormType=FORM_POST;
+            else if(strncasecmp(pTemp, "GET", 3) == 0) 
+              _psSessionData->nFormType=FORM_GET;
+            else 
+              _psSessionData->nFormType=FORM_UNKNOWN;
+
+            pTemp = strtok_r(NULL, "&", &pStrtokSavePtr);
+            if (pTemp != NULL)
+            {
+              _psSessionData->szFormUser = malloc( strlen(pTemp) + 1 );
+              memset(_psSessionData->szFormUser, 0, strlen(pTemp) + 1);
+              strncpy(_psSessionData->szFormUser, pTemp, strlen(pTemp));
+            }
+
+            pTemp = strtok_r(NULL, "&", &pStrtokSavePtr);
+            if (pTemp != NULL)
+            {
+              _psSessionData->szFormPass = malloc( strlen(pTemp) + 1);
+              memset(_psSessionData->szFormPass, 0, strlen(pTemp) + 1);
+              strncpy(_psSessionData->szFormPass, pTemp, strlen(pTemp));
+            }
+
+            pTemp = strtok_r(NULL, "", &pStrtokSavePtr);
+            if (pTemp != NULL)
+            {
+              _psSessionData->szFormRest = malloc( strlen(pTemp) + 1 );
+              memset(_psSessionData->szFormRest, 0, strlen(pTemp) + 1);
+              strncpy(_psSessionData->szFormRest, pTemp, strlen(pTemp));
+            }
+          }
 
           writeError(ERR_DEBUG_MODULE, "[%s] User-supplied Form User Field: %s", MODULE_NAME, _psSessionData->szFormUser);
           writeError(ERR_DEBUG_MODULE, "[%s] User-supplied Form Pass Field: %s", MODULE_NAME, _psSessionData->szFormPass);
@@ -393,10 +417,9 @@ int initHttp(_MODULE_DATA *_psSessionData, sLogin* login)
   FREE(_psSessionData->szUserAgent);
   FREE(_psSessionData->szDenySignal);
   FREE(_psSessionData->szFormData);
-  /* Ignore since they may point to data within szFormData (strtok) */ 
-  /* FREE(_psSessionData->szFormRest); */
-  /* FREE(_psSessionData->szFormUser); */
-  /* FREE(_psSessionData->szFormPass); */
+  FREE(_psSessionData->szFormRest);
+  FREE(_psSessionData->szFormUser);
+  FREE(_psSessionData->szFormPass); 
   FREE(_psSessionData);
 
   login->iStatus = LOGIN_DONE;
@@ -419,9 +442,13 @@ int sendPost(int hSocket, _MODULE_DATA* _psSessionData, char* szLogin, char* szP
   int nFormBufferSize = 0;
   int nRet = SUCCESS;
 
-  nFormBufferSize = asprintf(&bufForm, "%s%s&%s%s&%s", _psSessionData->szFormUser, szLogin, _psSessionData->szFormPass, szPassword, _psSessionData->szFormRest); 
+  if ((_psSessionData->szFormRest == NULL) || (_psSessionData->szFormRest[0] == 0))
+    nFormBufferSize = asprintf(&bufForm, "%s%s&%s%s", _psSessionData->szFormUser, szLogin, _psSessionData->szFormPass, szPassword); 
+  else
+    nFormBufferSize = asprintf(&bufForm, "%s%s&%s%s&%s", _psSessionData->szFormUser, szLogin, _psSessionData->szFormPass, szPassword, _psSessionData->szFormRest); 
+
   nSendBufferSize = asprintf(&bufSend, "POST /%s HTTP/1.1\r\nHost: %s\r\nUser-Agent: %s\r\nConnection: close\r\nContent-Type: application/x-www-form-urlencoded\r\nContent-Length: %i\r\n\r\n%s", _psSessionData->szDir, _psSessionData->szHostHeader, _psSessionData->szUserAgent, nFormBufferSize, bufForm);
-  
+
   if (medusaSend(hSocket, bufSend, nSendBufferSize, 0) < 0)
   {
     writeError(ERR_ERROR, "%s failed: medusaSend was not successful", MODULE_NAME);
@@ -439,7 +466,10 @@ int sendGet(int hSocket, _MODULE_DATA* _psSessionData, char* szLogin, char* szPa
   int nSendBufferSize = 0;
   int nRet = SUCCESS;
 
-  nSendBufferSize = asprintf(&bufSend, "GET /%s?%s%s&%s%s&%s HTTP/1.1\r\nHost: %s\r\nUser-Agent: %s\r\nConnection: close\r\n\r\n", _psSessionData->szDir, _psSessionData->szFormUser, szLogin, _psSessionData->szFormPass, szPassword, _psSessionData->szFormRest, _psSessionData->szHostHeader, _psSessionData->szUserAgent);
+  if ((_psSessionData->szFormRest == NULL) || (_psSessionData->szFormRest[0] == 0))
+    nSendBufferSize = asprintf(&bufSend, "GET /%s?%s%s&%s%s HTTP/1.1\r\nHost: %s\r\nUser-Agent: %s\r\nConnection: close\r\n\r\n", _psSessionData->szDir, _psSessionData->szFormUser, szLogin, _psSessionData->szFormPass, szPassword, _psSessionData->szHostHeader, _psSessionData->szUserAgent);
+  else
+    nSendBufferSize = asprintf(&bufSend, "GET /%s?%s%s&%s%s&%s HTTP/1.1\r\nHost: %s\r\nUser-Agent: %s\r\nConnection: close\r\n\r\n", _psSessionData->szDir, _psSessionData->szFormUser, szLogin, _psSessionData->szFormPass, szPassword, _psSessionData->szFormRest, _psSessionData->szHostHeader, _psSessionData->szUserAgent);
   
   if (medusaSend(hSocket, bufSend, nSendBufferSize, 0) < 0)
   {
@@ -493,7 +523,7 @@ int tryLogin(int hSocket, _MODULE_DATA* _psSessionData, sLogin** login, int nPor
   pTemp = ((char*)index(pReceiveBuffer, ' ')) + 1;
   if (strncmp(pTemp, "200 OK", 6) != 0 )
   {
-    writeError(VB_GENERAL, "The answer was NOT successfully received, understood, and accepted: error code was %.6s", pTemp);
+    writeError(ERR_ERROR, "The answer was NOT successfully received, understood, and accepted: error code %.4s", pTemp);
     (*login)->iResult = LOGIN_RESULT_UNKNOWN;
     setPassResult(*login, szPassword);
     return HSTATE_RUNNING;
